#ifndef APPLICATION_H
#define APPLICATION_H

#ifdef Q_OS_WIN
#pragma comment(lib, "crypt32.lib")
#endif

#include <QApplication>
#include <QTemporaryFile>
#include <QObject>
#include <QString>
#include <QStringList>
#include <QIcon>
#include <QPixmap>
#include <QFileInfo>
#include <QFile>
#include <QMessageBox>
#include <QMimeDatabase>
#include <QHash>
#include <QProcess>
#include <QCryptographicHash>
#include <QByteArray>
#include <QFileSystemWatcher>
#include <QTemporaryDir>

#include "constants.h"
#include "settings.h"
#include "cupsmanager.h"
#include "lwclient.h"
#include "initprogress.h"

class Application : public QObject
{
	Q_OBJECT

public:
	Application(const QString & id);
	Application( Application * app );

	void setFile(QString file) { this->file = file; }
	void setId(QString id) { this->id = id; } // Id
	void setName(QString name) { this->name = name; } // Name
	void setDescription(QString description) { this->description = description; } // Description
	void setCommand(QString command) { this->command = command; }
	void setServer(QString server) { this->server = server; }
	void setServerId(QString serverId) { this->serverId = serverId; }
	void setOS(int os) { this->os = os; }
	void setIcon(QIcon icon) { this->icon = icon; }
	void setArguments(QStringList arguments) { this->arguments = arguments; }
	void setWorkingPath(QString workingPath) { this->workingPath = workingPath; }
	void setGroups(QStringList groups) { this->groups = groups; }
	void setMimeTypes(QStringList mimeTypes) { this->mimeTypes = mimeTypes; }
	void setProtocol(QString protocol) { this->protocol = protocol; }
	
	QString getId() { return id.trimmed(); } // Id
	QString getName() { return name.trimmed(); } // Name
	QString getDescription() { return description.trimmed(); } // Description
	QString getCommand() { return command.trimmed(); }
	QString getServer() { return server.trimmed(); }
	QString getServerId() { return serverId.trimmed(); }
	int getOS() { return os; }
	QIcon getIcon() { return icon; }
	QStringList getArguments() { return arguments; }
	QString getWorkingPath() { return workingPath; }
	QString getFile() { return file; }
	QStringList getGroups() { return groups; }
	QStringList getMimeTypes() { return mimeTypes; }
	QString getProtocol() { return protocol; }
	Q_PID getPid() { return pid; }

	void launch(QStringList stdInput = QStringList());

	static void setFileManager();
	static Application * getFileManager();
	static void launchFileManager( QString path );

	static void deleteList( QList<Application*> & list );
	static void updateApplicationCache(InitProgress * initProgress = nullptr);
	static void deleteApplicationCache();

	static QString mimeType(QFileInfo fileInfo);
	static QIcon getIcon(QFileInfo file);
	static QIcon findIcon(QString iconName);
	static Application * preferredProgram(QString extension);
	static QList<Application*> applicationList(QString extension = QString());
	static QList<Application*> applicationDeletedList(QString extension);
	static QList<Application*> applicationListByGroup(QString group);
	static QList<Application*> favoriteApplicationList();
	static QStringList applicationGroupsList();
	static QStringList applicationGroupIdsList();
	static QIcon applicationGroupIcon(QString groupId);
	static QStringList getExtensionsList();
	static QStringList getValidApplications();
	static void addExtension(QString extension);
	static Application * openFile(QString file, QString application = QString());

	static QString windowsPath(QString file);
	static QStringList getArguments(QString command, QStringList arguments, QString file, int os, QString serverAddress );
	static QStringList remoteApplicationArguments(QString command, QStringList args, QString file, QString serverAddress );
	static QString resolveWindowsEnvironment( QString command );
	static bool isRemote(QString server, int os);
	static QString checkCommand(QString command, QString server, int os, QString protocol);
	static bool isValid(QString command, QString server, int os, QString protocol);
	static QStringList mstscArguments(QString command, QStringList args, QString file, quint16 port, QString serverAddress);
	static QString cryptPassword(QString password);
	static void clearMstsc(QString tempFile);
	static QStringList parseArguments( QString args );

public slots:
	void deleteAppInstance();

signals:
	void finished(int, QProcess::ExitStatus);
	void progress(int);

private slots:
	void finish( int exitCode, QProcess::ExitStatus exitStatus );
	void watcherFileChanged( QString fileName );

private:
	static void setLocalValues(QString id, Application * app);
	static void setServerValues(QString id, Application * app);
	static QString firstString( QStringList stringList );
	static Application * fromDesktopFile(QString desktopFile, QStringList & appNameList);

	QString id;
	QString name;
	QString description;
	QString command;
	QString workingPath;
	int os;
	QString server;
	QString serverId;
	QStringList arguments;
	QIcon icon;
	QString file;
	QStringList groups;
	QStringList mimeTypes;
	QString protocol;
	QFileSystemWatcher fileSystemWatcher;

	Q_PID pid;

	static Application * fileManager;
	static QStringList validApplications;
	static QStringList groupCache;
	static QStringList groupIdCache;
	static QList< Application* > appList;
	static QHash< QString,QList<Application*> > appCache; // appCache[<extension>] = List of applications for this extension
	static QHash< QString,QByteArray > groupIconCache; // groupIconCache[<group>] = Icon of group
	static QHash< QString,QList<Application*> > appByGroupCache; // appByGroupCache[<group>] = List of applications for this group
	static QHash< QString,QList<Application*> > groupByAppCache; // groupByAppCache[<group>] = List of groups for this application
	static QHash< QString,QStringList > extensionsCache; // extensionsCache[<mime type>] = List of extensions for this mime type

	static QString socketDir;
};

#endif
