#include "placeswidget.h"
#include "manager.h"

#include <QDir>

AppsWidget::AppsWidget(QWidget * parent) : IconView( parent, true )
{
	setFileSystemMode( false );
	init();
	updateApps();
}

void AppsWidget::init()
{
	model = new AppsModel;
	setModel( model );
	setEditTriggers(QAbstractItemView::NoEditTriggers);
	setStyleSheet(QString(	"QListView {"
								"	border-style: none;"
								"	background-color: transparent;"
								"	color: " FONTCOLOR ";"
							"}"));
	setResizeMode( QListView::Fixed );
	setMovement( QListView::Snap );
	setDragDropMode(QAbstractItemView::InternalMove);
	setOrientation( Qt::Horizontal );

	connect( this, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(doubleClicked(QModelIndex)));
	connect( this, SIGNAL(clicked(QModelIndex)), this, SLOT(doubleClicked(QModelIndex)));
	connect( this, SIGNAL(indexesMoved(QModelIndexList)), this, SLOT(saveIconPositions()));
}

void AppsWidget::updateApps()
{
	model->clear();
	// Add applications
	QList<Application*> validApplications = Application::favoriteApplicationList();
	int count = validApplications.count();
	for (int i = 0; i < count; i++)
	{
		addApplication(validApplications[i]);
	}
}

void AppsWidget::addApplication(Application * app)
{
	model->appendItem( app );
}

void AppsWidget::doubleClicked(QModelIndex index)
{
	QStandardItem * item = model->itemFromIndex( index );
	if ( index.isValid() )
	{
		Application * app = model->getItem( item );
		if ( app != 0 )
		{
			emit launchApplication( app );
		}
	}
}

void AppsWidget::loadIconPositions()
{
	QSettings settings(USER_CONFIG, QSettings::IniFormat);
	settings.beginGroup("Desktop");
	QStringList groupList = settings.allKeys();
	for ( int i = 0; i < groupList.count(); i++ )
	{
		if ( ! groupList[i].startsWith("AppsWidget.") )
		{
			continue;
		}
		QString name = groupList.at(i).section( "AppsWidget.", 1, 1 );
		qApp->processEvents();
		Application * app = 0;
		int row = -1;
		int count = model->rowCount();
		QModelIndex index;
		for ( int k = 0; k < count; k++ )
		{
			index = model->index( k, 0 );
			if ( index.isValid() )
			{
				app = model->getItem( model->itemFromIndex(index) );
				if ( app != 0 )
				{
					if ( app->getName() == name )
					{
						row = k;
						break;
					}
				}
			}
		}
		if ( app == 0 || row < 0 || ! index.isValid() )
		{
			qDebug("Removing non existent app %s", name.toUtf8().data());
			settings.remove( "AppsWidget." + name );
			continue;
		}

		QPoint pos = settings.value( groupList[i] ).toPoint();
		QModelIndex existent = this->indexAt( pos );
		QStandardItem * existentItem = model->itemFromIndex( existent );
		if ( existent.isValid() )
		{
			Application * existentApp = model->getItem( existentItem );
			if ( existentApp != 0 )
			{
				if ( existentApp->getName() == name ) // If it is itself, everything is ok, so continue loop
				{
					continue;
				} else {
					swapPosition( index, existent );
				}
			} else {
				qDebug("Rare problem. Not valid app?");
				continue;
			}
		} else {
			setPositionForIndex( pos, model->index( row, 0 ) );
		}
	}
	settings.endGroup();
	settings.sync();
}

void AppsWidget::saveIconPositions()
{
	QSettings settings(USER_CONFIG, QSettings::IniFormat);
	settings.beginGroup("Desktop");
	for ( int i = 0; i < model->rowCount(); i++ )
	{
		QModelIndex index =  model->index(i, 0);
		if ( index.isValid() )
		{
			Application * app = model->getItem( model->itemFromIndex(index) );
			if ( app != 0 )
			{
				QString name = app->getName();
				QPoint pos = rectForIndex( index ).topLeft();
				settings.setValue( "AppsWidget." + name, QVariant(pos) );
			}
		}
	}
	settings.endGroup();
	settings.sync();
}

void AppsWidget::dropEvent(QDropEvent* e)
{
	QPoint pos = e->pos();
	QModelIndex target = this->indexAt( pos );
	QString data = QUrl::fromPercentEncoding( e->mimeData()->data( mimeFormat ) );

	if ( e->source() == (QObject*) this )
	{
		if ( target.isValid() )
		{
			QStringList dataList = data.split("\n");
			if ( dataList.count() > 0 )
			{
				QList<QModelIndex> indexes = selectedIndexes();
				if ( indexes.count() > 0 )
				{
					QModelIndex sourceIndex = indexes[0]; // By default only one item at a time
					if ( sourceIndex.isValid() )
					{
						QModelIndexList list;
						swapPosition( target, indexes[0] );
						list += target;
						list += indexes[0];
						emit indexesMoved( list );
					}
				}
			}
		} else {
			QModelIndexList indexes = selectedIndexes();
			emit indexesMoved( indexes );
			QListView::dropEvent(e);
		}
	} else {
		e->ignore();
	}
}
