#include "cloudconfigwizard.h"

CloudConfigWizard::CloudConfigWizard(QStringList existentConfigs, QWidget *parent) : QWizard(parent)
{
	this->existentConfigs = existentConfigs;

	addPage( createInitPage() );
	addPage( createWebViewPage() );
	addPage( createFinishPage() );
	setWindowTitle( tr("Cloud drive configuration wizard") );

	connect(this, SIGNAL(currentIdChanged(int)), this, SLOT(pageChanged(int)));
}

void CloudConfigWizard::reject()
{
	if ( LWClient::instance()->isConnected() )
	{
		LWClient::instance()->sendMessage("cancel_set_cloud_drive");
	}
	QWizard::reject();
}

CloudConfigWizardPage *CloudConfigWizard::createInitPage()
{
	QGridLayout * layout = new QGridLayout;
	CloudConfigWizardPage * page = new CloudConfigWizardPage;
	page->setTitle( tr("Introduction") );
	page->setValid( true );

	QLabel * introLabel = new QLabel( tr("Welcome to the cloud drive configuration wizard") );

	QLabel * nameLabel = new QLabel( tr("Connection name:") );
	nameLineEdit = new QLineEdit( "" );

	QLabel * cloudTypeLabel = new QLabel( tr("Cloud disk type:") );
	cloudTypeComboBox = new QComboBox;
	//cloudTypeComboBox->addItem(QIcon(":/icons/applications/dropbox.png"), "Dropbox", "dropbox");
	cloudTypeComboBox->addItem(QIcon(":/icons/applications/googledrive.png"), "Google Drive", "googledrive");

	layout->addWidget( introLabel, 0, 0, 1, 2);
	layout->addWidget( nameLabel, 1, 0);
	layout->addWidget( nameLineEdit, 1, 1);
	layout->addWidget( cloudTypeLabel, 2, 0);
	layout->addWidget( cloudTypeComboBox, 2, 1);
	page->setLayout( layout );

	return page;
}

CloudConfigWizardPage * CloudConfigWizard::createWebViewPage()
{
	QVBoxLayout * layout = new QVBoxLayout;
	CloudConfigWizardPage * page = new CloudConfigWizardPage;
	page->setTitle( tr("Cloud configuration") );
#if QT_VERSION >= QT_VERSION_CHECK(5, 7, 0)
	QWebEngineProfile::defaultProfile()->setHttpAcceptLanguage(Settings::getUserValue("Language", "en").toString());
	webView = new QWebEngineView;
#else
	webView = new QWebView;
#endif
	layout->addWidget( webView );
	page->setLayout( layout );
	page->setCommitPage( true );

	return page;
}

CloudConfigWizardPage *CloudConfigWizard::createFinishPage()
{
	QGridLayout * layout = new QGridLayout;
	CloudConfigWizardPage * page = new CloudConfigWizardPage;
	page->setTitle( tr("Finish wizard") );
	page->setValid( true );

	finishLabel = new QLabel();
	layout->addWidget( finishLabel, 0, 0);
	page->setLayout( layout );

	return page;
}

void CloudConfigWizard::nameChanged(QString name)
{
	if ( name.isEmpty() || existentConfigs.contains(name, Qt::CaseInsensitive) )
	{
		button(QWizard::NextButton)->setEnabled( false );
	} else {
		button(QWizard::NextButton)->setEnabled( true );
	}
}

void CloudConfigWizard::pageChanged(int id)
{
	bool accept = true;
	Settings::logInfo(QString("CloudConfig wizard page changed to %1").arg(id));
	switch (id)
	{
	case 0:
		break;
	case 1:
		if (nameLineEdit->text().isEmpty())
		{
			accept = false;
			this->back();
			QMessageBox::critical( this, tr("Empty connection name"), tr("The connection name must not be empty. Please, set a connection name."));
		} else {
			if ( existentConfigs.contains(nameLineEdit->text(), Qt::CaseInsensitive) )
			{
				accept = false;
				this->back();
				QMessageBox::critical( this, tr("Duplicate connection name"), tr("The connection name you have selected already exists. Please, select a new name."));
			}
		}
		if ( accept )
		{
			setCursor( Qt::WaitCursor );
			Settings::logInfo(QString("Adding %1 account '%2'").arg(nameLineEdit->text()).arg(cloudTypeComboBox->currentData().toString()));
			QString url = addAccount( nameLineEdit->text(), cloudTypeComboBox->currentData().toString() );
			if ( url.startsWith("Error:") )
			{
				setCursor( Qt::ArrowCursor );
				QMessageBox::critical( this, tr("Error adding account"), tr("Cannot add account. System returned this error:") + "\n" + url);
			} else {
				if ( url.simplified().isEmpty() )
				{
					QMessageBox::critical( this, tr("Error adding account"), tr("Cannot add account. Librework server returned an empty URL. Please, contact system administrator"));
					this->back();
					setCursor( Qt::ArrowCursor );
				} else {
					Settings::logInfo(QString("Loading web page %1...").arg(url));
					webView->load( QUrl( url ) );
					connect( webView, SIGNAL(loadFinished(bool)), this, SLOT(webViewLoaded(bool)));
					connect( webView, SIGNAL(loadStarted()), this, SLOT(webViewLoadStarted()));
				}
			}
		}
		break;
	case 2:
		break;
	default:
		break;
	}
}

QString CloudConfigWizard::addAccount(QString accountName, QString cloudType)
{
	QString url;
	if ( LWClient::instance()->isConnected() )
	{
		Settings::logInfo(QString("Start setting %1 cloud drive %2").arg(cloudType).arg(accountName));
		url = LWClient::instance()->sendMessage("set_cloud_drive?" + cloudType + "&" + accountName);
	} else {
		QMessageBox::critical( this, tr("Error adding account"), tr("Cannot add account. Cannot connect to Librework Server."));
		Settings::logError(QString("Cannot add %1 account %2").arg(cloudType).arg(accountName));
		this->back();
	}
	qDebug("url: %s", url.toUtf8().data());
	return url;
}

void CloudConfigWizard::webViewLoaded(bool)
{
	if (webView->title().startsWith("Denied error=") )
	{
			reject();
	}

#if QT_VERSION >= QT_VERSION_CHECK(5, 7, 0)
	webView->page()->toHtml(
				[this] (const QString &result)
				{
					QString code;
					if ( result.contains("auth-code") ) // Dropbox
					{
						int from = result.indexOf("data-token=");
						from = result.indexOf("\"", from) + 1;
						int to = result.indexOf("\"", from);
						code = result.mid( from, to - from );
					}
					if ( result.contains("<input id=\"code\"") ) // Google Drive
					{
						int from = result.indexOf("<input id=\"code\"");
						from = result.indexOf("value=", from);
						from = result.indexOf("\"", from) + 1;
						int to = result.indexOf("\"", from);
						code = result.mid( from, to - from );
					}
					this->setCursor( Qt::ArrowCursor );
					if ( ! code.isEmpty() )
					{
						QString ret = LWClient::instance()->sendMessage("finish_set_cloud_drive?" + code);
						if ( ret.startsWith("Error"))
						{
							finishLabel->setText( tr("Something failed adding cloud account") );
						} else {
							finishLabel->setText( tr("Congratulations! Your cloud account has been configured") );
						}
						CloudConfigWizardPage * page = (CloudConfigWizardPage *)this->currentPage();
						page->setValid( true );
						this->next();
					}
				});
#else
	QString title = webView->title();
	QString code;

	if (title.startsWith("Success code=") ) // Google Drive
	{
		QWebElement element = webView->page()->currentFrame()->documentElement();
		QWebElementCollection allInputs = element.findAll("input");

		for (int i = 0; i < allInputs.count(); i++)
		{
			if (allInputs.at(i).attribute("id") == "code")
			{
				code = allInputs[i].attribute("value", QString());
			}
		}
	} else { // Dropbox
		QWebElement element = webView->page()->currentFrame()->documentElement();
		QWebElementCollection allInputs = element.findAll("span");

		for (int i = 0; i < allInputs.count(); i++)
		{
			if (allInputs.at(i).attribute("class") == "auth-code")
			{
				code = allInputs[i].toPlainText();
			}
		}
	}

	this->setCursor( Qt::ArrowCursor );
	if ( ! code.isEmpty() )
	{
		QString ret = LWClient::sendMessage("finish_set_cloud_drive?" + code, serverAddress, serverPort);
		if ( ret.startsWith("Error"))
		{
			finishLabel->setText( tr("Something failed adding cloud account") );
		} else {
			finishLabel->setText( tr("Congratulations! Your cloud account has been configured") );
		}
		CloudConfigWizardPage * page = (CloudConfigWizardPage *)this->currentPage();
		page->setValid( true );
		this->next();
	}
#endif
}

void CloudConfigWizard::webViewLoadStarted()
{
	setCursor(Qt::WaitCursor);
}
