#include "desktop.h"
#include "manager.h"

#include <QDir>

Desktop::Desktop(InitProgress *initProgress, QWidget * parent, QString color) : PixmapFrame( parent )
{
	QRect geometry = QApplication::primaryScreen()->geometry();
	QRect desktopGeometry( geometry.x(), geometry.y(), geometry.width(), geometry.height() - TAB_BUTTON_HEIGHT );
	setGeometry( desktopGeometry );
	setFixedSize( desktopGeometry.size() );

	setWindowFlags( Qt::CustomizeWindowHint | Qt::FramelessWindowHint );
	WinAction::setDesktopWindow( (HWND)this->winId() );

	setAttribute(Qt::WA_ShowWithoutActivating, true);
	setWindowTitle(tr("Librework Desktop"));

	this->initProgress = initProgress;

	WinAction::setRootWindowBackgroundColor( QColor( BACKGROUNDCOLOR ) );

	QVBoxLayout * layout = new QVBoxLayout;
	layout->setContentsMargins(0, 0, 0, 0);
	layout->setSpacing( 0 );

	placesWidget = new PlacesWidget;
	placesWidget->setFixedHeight(LISTVIEWITEM_GRID_SIZE.height());
	placesWidget->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Fixed);
	layout->addWidget( placesWidget );

	QHBoxLayout * hLineLayout = new QHBoxLayout;
	hLineLayout->setSpacing( 0 );
	hLineLayout->setContentsMargins( 0, 0, 0, 0 );
	QSpacerItem * hLineLeftSpacer = new QSpacerItem( SEPARATOR_MARGIN, 0 );
	hLineLayout->addItem( hLineLeftSpacer );
	hLine = new QWidget;
	hLine->setFixedHeight(SEPARATOR_WIDTH);
	hLine->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
	hLine->setStyleSheet( QString("QWidget{ background-color: rgba(255, 255, 255, %1); };").arg(SEPARATOR_ALPHA));
	hLineLayout->addWidget( hLine );
	QSpacerItem * hLineRightSpacer = new QSpacerItem( SEPARATOR_MARGIN, 0 );
	hLineLayout->addItem( hLineRightSpacer );

	layout->addLayout( hLineLayout );

	spacerHeight = 10;
	spacer = new QSpacerItem(0, spacerHeight);
	layout->addItem( spacer );

	splitter = new Splitter;
	splitter->setOrientation( Qt::Horizontal );
	splitter->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding );
	QPalette palette = splitter->palette();
	palette.setBrush(QPalette::Base, Qt::transparent);
	splitter->setPalette(palette);
	splitter->setAttribute(Qt::WA_OpaquePaintEvent, false);
	splitter->setAttribute(Qt::WA_TranslucentBackground);
	splitter->setAttribute( Qt::WA_NoSystemBackground );
	connect( splitter, SIGNAL(splitterMoved(int,int)), this, SLOT(saveGeometry()));

	setLayout( layout );

	qApp->processEvents();

	setStyleSheet( QString("background-color: %1").arg(color) );

	setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
	setContentsMargins( 10, 5, 10, 10 );
	splitter->setContentsMargins( 0, 0, 0, 0 );

	appsWidget = new AppsWidget;
	appsWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
	splitter->addWidget( appsWidget );
	connect( appsWidget, SIGNAL(launchApplication(Application*)), this, SIGNAL(launchApplication(Application*)));

	folderView = new FileSystemWidget;
	folderView->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
	splitter->addWidget( folderView );

	layout->addWidget( splitter );

	useDesktopWidgets = Settings::getUserValue("General/DesktopWidgets", false).toBool();

	if ( useDesktopWidgets )
	{
		widgetsWidget = new TableWidget;
		widgetsScrollArea = new QScrollArea;
		widgetsScrollArea->setStyleSheet("QScrollArea {border-style: none;}");
		widgetsScrollArea->setAttribute(Qt::WA_TranslucentBackground);
		widgetsScrollArea->setWidget( widgetsWidget );
		widgetsScrollArea->setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
		widgetsScrollArea->setAlignment( Qt::AlignTop | Qt::AlignLeft);
		splitter->addWidget( widgetsScrollArea );
		/*connect( widgetsWidget, SIGNAL(mousePressed()), this, SLOT(mousePressed()));*/
	}

	splitter->setHandleWidth( 32 );

	/*connect( appsWidget, SIGNAL(mousePressed()), this, SLOT(mousePressed()));
	connect( folderView, SIGNAL(mousePressed()), this, SLOT(mousePressed()));*/
}

void Desktop::updateDesktopWidgets()
{
	// Add widgets
	for ( int i = 0; i < desktopWidgets.count(); i++ )
	{
		locateWidget(i);
	}
}

void Desktop::reloadFavorites()
{
	resize();
}

void Desktop::resize( QSize size )
{
	QSize newSize = size;
	if ( size == QSize(0, 0))
	{
		newSize = Manager::getDesktopSize();
	}

	setFixedSize( newSize );
	PixmapFrame::resize();

	splitter->setFixedHeight( newSize.height() - hLine->height() - spacerHeight - contentsMargins().bottom() - placesWidget->height() );
	appsWidget->setFixedHeight( splitter->height() );

	appsWidget->updateApps();
	appsWidget->loadIconPositions();

	placesWidget->loadIconPositions();

	if ( useDesktopWidgets )
	{
		updateDesktopWidgets();
	}

	if ( useDesktopWidgets )
	{
		widgetsScrollArea->resize( widgetsWidget->size() );
	}

	QString resolution = QString::number( newSize.width() ) + "x" + QString::number( newSize.height() );
	QStringList splitterGeometry = Settings::getUserValue("WindowGeometry/Desktop.splitter-" + resolution, QVariant() ).toStringList();
	if ( splitterGeometry.count() > 0 )
	{
		QList<int> sizes;
		for ( int i = 0; i < splitterGeometry.count(); i++ )
		{
			sizes += splitterGeometry.at(i).toInt();
		}
		splitter->setSizes( sizes );
	}

	if ( initProgress != nullptr )
	{
		initProgress->close();
		delete initProgress;
		initProgress = nullptr;
	}
}

void Desktop::saveGeometry()
{
	QSize newSize = Manager::getDesktopSize();
	QStringList splitterGeometry;
	QList<int> sizes = splitter->sizes();
	for ( int i = 0; i < sizes.count(); i++ )
	{
		splitterGeometry += QString::number( sizes.at(i) );
	}
	QString resolution = QString::number( newSize.width() ) + "x" + QString::number( newSize.height() );
	Settings::setUserValue("WindowGeometry/Desktop.splitter-" + resolution, QVariant(splitterGeometry));
}

void Desktop::closeEvent(QCloseEvent *closeEvent)
{
	closeEvent->ignore();
}

HWND Desktop::getWindowId()
{
	return (HWND)0; // Otherwise, if return winId() the tab is closed because a destroyNotify is emitted. See manager.cpp
}

void Desktop::initializeDesktopWidgets()
{
	QDir widgetsDir(SHAREDIR "widgets");
	QStringList entries = widgetsDir.entryList(QStringList("*.ini"));
	for (int i = 0; i < entries.count(); i++)
	{
		addDesktopWidget( SHAREDIR "widgets" + QString(QDir::separator()) + entries[i] );
	}
}

void Desktop::addDesktopWidget(QString iniFile)
{
	QSettings widgetSettings( iniFile, QSettings::IniFormat );

	QString name = widgetSettings.value("Librework Widget/Name").toString();
	QString command = "\"" + QString(SHAREDIR) + "/"  + widgetSettings.value("Librework Widget/Command").toString() + "\"";
	int minWidth = widgetSettings.value("Librework Widget/MinimumWidth").toInt();
	int minHeight = widgetSettings.value("Librework Widget/MinimumHeight").toInt();
	int maxWidth = widgetSettings.value("Librework Widget/MaximumWidth").toInt();
	int maxHeight = widgetSettings.value("Librework Widget/MaximumHeight").toInt();
	QString verticalAlignment = widgetSettings.value("Librework Widget/VerticalAlignment", "top").toString().toLower();
	QString horizontalAlignment = widgetSettings.value("Librework Widget/HorizontalAlignment", "right").toString().toLower();

	QProcess * process = new QProcess();
	process->start( command );
	process->waitForStarted();
	connect( process, SIGNAL(finished(int)), process, SLOT(deleteLater()));
	connect( this, SIGNAL(destroyed(QObject*)), process, SLOT(kill()));
	qDebug("Launching %s (PID: %d)", command.toUtf8().data(), (int)process->processId());
	process->waitForStarted();

	//QSharedMemory * sharedMemory = new QSharedMemory("Librework " + name);

	HWND winId = (HWND) -1;
	int retries = 25;
	bool success = false;
	while (retries > 0 && success == false)
	{
		QThread::msleep(100);
		/*if ( sharedMemory->attach() )
		{
			success = true;
			sharedMemory->lock();
			winId = QString( (char*)sharedMemory->data()).toLongLong();
			sharedMemory->unlock();
			sharedMemory->detach();
			delete sharedMemory;
			sharedMemory = 0;
		}*/
		if ( winId == (HWND)-1 )
		{
			HWND hWnd = WinInfo::getWindow( process );
			if ( hWnd != (HWND)-1 )
			{
				success = true;
				winId = hWnd;
			}
		}
		retries--;
	}

	/*if ( sharedMemory != 0 )
	{
		delete sharedMemory;
	}*/

	if ( success )
	{
		qDebug("Successfuly exited loop after %d retries of %d", 25 - retries, 25);
		qDebug("Attaching widget window %llu to deskmenu...", (long long) winId);
		QWidget * container = new QWidget;

		QVBoxLayout * layout = new QVBoxLayout;
		layout->setContentsMargins(0, 0, 0, 0);
		QWindow * containerWindow = QWindow::fromWinId( (WId) winId );
		QWidget * containerWidget = QWidget::createWindowContainer( containerWindow );
		containerWidget->setMaximumWidth(maxWidth);
		containerWidget->setMinimumHeight(minWidth);
		containerWidget->setMaximumHeight(maxHeight);
		containerWidget->setMinimumHeight(minHeight);
		container->setFixedSize(minWidth, minHeight);
		layout->addWidget( containerWidget );
		container->setLayout( layout );

		QGraphicsDropShadowEffect * effect = new QGraphicsDropShadowEffect();
		effect->setBlurRadius( 20 );
		effect->setXOffset( 8 );
		effect->setYOffset( 8 );
		QColor color = QColor(0, 0, 0);
		color.setAlpha( 150 );
		effect->setColor( color );
		container->setGraphicsEffect( effect );

		desktopWidgets += container;
		desktopWidgetsMinSizes += QSize(minWidth, minHeight);
		desktopWidgetsMaxSizes += QSize(maxWidth, maxHeight);
		desktopWidgetsVerticalAlignment += verticalAlignment;
		desktopWidgetsHorizontalAlignment += horizontalAlignment;
	} else {
		qDebug("Cannot find widget window id. Killing process.");
		process->kill();
	}
}

void Desktop::locateWidget(int widgetId)
{
	qDebug("Locating widget %d", widgetId);
}

void Desktop::mousePressed()
{
	emit selectWidget( this );
}

void Desktop::setLogo()
{
	bool centered = true;
	QPixmap logo;
	if ( LWClient::instance()->isConnected() )
	{
		QByteArray ic = LWClient::instance()->sendMessageByteArray( "get_user_logo?" + Settings::getUserName() );

		logo.loadFromData( ic );
	}

	if ( logo.isNull() )
	{
		centered = false;
		logo = QPixmap(ICON_DIR_OTHER + QString(QDir::separator()) + DEFAULT_WALLPAPER);
	}
	setOpacity( 0.98 );
	setPixmap( logo, centered, false );
}
