#include "filesystemwidget.h"
#include "ui_filesystemwidget.h"
#include "manager.h"

FileSystemWidget::FileSystemWidget(QWidget *parent, bool standalone, QString openLocation) :
    ui(new Ui::FileSystemWidget)
{
	this->standalone = standalone;
	if ( standalone )
	{
		int serverPort = Settings::getValue( "Server/LibreworkServerPort" ).toInt();
		QString serverAddress = Settings::getValue( "Server/LibreworkServerAddress" );
		QString token;
		qDebug("Looking for token on file: %s", QString(TOKEN_FILE).toUtf8().data());
		if (QFile::exists( TOKEN_FILE ))
		{
			token = Settings::getValue( TOKEN_FILE, "UserToken" );
		}

		if ( serverPort > 0 && ! token.isEmpty() && ! serverAddress.isEmpty() )
		{
			if ( LWClient::instance()->connect( serverAddress, serverPort ) )
			{
				if ( LWClient::instance()->authenticate( Settings::getUserName(), token ) )
				{
					qDebug("Authentication ok");
				} else {
					qDebug("Authentication failed");
				}
			} else {
				qDebug("Cannot connect to Librework server");
			}
		}
		Application::updateApplicationCache();
		QWidget();
	} else {
		QWidget(parent, Qt::CustomizeWindowHint | Qt::FramelessWindowHint );
	}
	setWindowIcon(QIcon(":/icons/32x32/folder.png"));
	model = nullptr;
	searchProgress = nullptr;
	searching = false;
	cut = false;
	ui->setupUi(this);
	init( openLocation );
	static_cast<IconView*>(ui->iconView)->setFileSystemMode( true );
	static_cast<IconView*>(ui->iconView)->setDragDropMode( QAbstractItemView::DragDrop );
}

FileSystemWidget::FileSystemWidget(QIcon icon, QWidget * parent) :
	QWidget(parent, Qt::CustomizeWindowHint | Qt::WindowCloseButtonHint), ui(new Ui::FileSystemWidget)
{
	standalone = false;
	setWindowIcon(icon);
	model = nullptr;
	ui->setupUi(this);
	remoteClientManager = nullptr;

	ui->splitter->setStretchFactor(0, 1);
	ui->splitter->setStretchFactor(1, 3);

	ui->iconView->setSelectionMode(QAbstractItemView::NoSelection);
}

void FileSystemWidget::init(QString openLocation)
{
	remoteClientManager = new RemoteClientManager();
	if ( standalone )
	{
		setStyleSheet(QString(	"QWidget {"
								"	background-color: white;"
								"}"));
		QString listStyle = QString(	"QAbstractItemView {"
										"	background-color: white;"
										"	color: " FILESYSTEMWIDGETFONTCOLOR ";"
										"}");
		ui->iconView->setStyleSheet(listStyle);
		ui->listView->setStyleSheet(listStyle);
		ui->iconView->setSelectionRectVisible( true );
		static_cast<IconView*>(ui->iconView)->setSelectionShadow( true );
	} else {
		ui->iconView->setStyleSheet(QString(
									"QListView {"
									"	background-color: transparent;"
									"	color: " FONTCOLOR ";"
									"}"
									"QDialog {"
									"	background-color: white;"
									"	color: " FILESYSTEMWIDGETFONTCOLOR ";"
									"}"
									));
		static_cast<IconView*>(ui->iconView)->setPressedButtons( true );
		QTimer::singleShot( NETWORK_START_WAIT_TIME, remoteClientManager, SLOT(start()));
		connect( remoteClientManager, SIGNAL(update()), this, SLOT(updatePlaces()));
		connect( remoteClientManager, SIGNAL(update()), this, SLOT(setHome()));
		setWindowFlags(Qt::FramelessWindowHint);
	}

	fillPlaces();

	initUi(openLocation);

	setEnabledButtons();

	ui->fileInfo->setMinimumHeight( LIBREWORK_FONT_BASE_SIZE * 3 );
	ui->fileInfo->setAlignment(Qt::AlignTop);

	if ( standalone )
	{
		setWindowTitle(tr("Librework File Manager"));

		ui->iconView->setSelectionMode( QAbstractItemView::ExtendedSelection );
		ui->listView->setSelectionMode( QAbstractItemView::ExtendedSelection );

		setContentsMargins(7, 7, 7, 7);
		Settings::loadWidgetGeometry("FileSystemWidget", this);
		QStringList splitterGeometry = Settings::getUserValue("WindowGeometry/FileSystemWidget.splitter", QVariant() ).toStringList();
		if ( splitterGeometry.count() > 0 )
		{
			QList<int> sizes;
			for ( int i = 0; i < splitterGeometry.count(); i++ )
			{
				sizes += splitterGeometry.at(i).toInt();
			}
			ui->splitter->setSizes( sizes );
		}
		QStringList listViewGeometry = Settings::getUserValue("WindowGeometry/FileSystemWidget.listView", QVariant() ).toStringList();
		if ( listViewGeometry.count() == 0 )
		{
			listViewGeometry.append("250"); // minimum width for file name column
		}
		for ( int i = 0; i < listViewGeometry.count(); i++ )
		{
			ui->listView->setColumnWidth( i, listViewGeometry[i].toInt() );
		}
		QString viewType = Settings::getUserValue("WindowGeometry/FileSystemWidget.viewType", QVariant("icon") ).toString();
		if ( viewType.toLower() == "list" )
		{
			changeView();
			ui->iconViewButton->setEnabled( true );
			ui->listViewButton->setEnabled( false );
			ui->iconView->hide();
			ui->listView->show();
			currentView = ui->listView;
		} else {
			currentView = ui->iconView;
		}
	} else {
		setWindowTitle(tr("Librework Files"));
		setContentsMargins(0, 0, 0, 0);
		ui->splitter->setContentsMargins(0, 0, 0, 0);
		ui->verticalLayout->setContentsMargins(0, 0, 0, 0);
		ui->verticalLayout->setSpacing( 0 );
		currentView = ui->iconView;

		ui->placesView->hide();
		ui->lineSplitter->hide();

		ui->addressLineEdit->hide();
		ui->copyButton->hide();
		ui->removeButton->hide();
		ui->pasteButton->hide();
		ui->newFolderButton->hide();
		ui->cutButton->hide();
		ui->lineTopBar->hide();
		ui->upButton->hide();
		ui->renameButton->hide();
		ui->iconViewButton->hide();
		ui->listViewButton->hide();

		ui->lineFileInfo->hide();
		ui->fileInfo->hide();

		ui->lineTop->hide();

		ui->iconView->setFlow( QListView::TopToBottom );

		QPalette palette = this->palette();
		palette.setBrush(QPalette::Base, Qt::transparent);
		setPalette(palette);
		setAttribute(Qt::WA_OpaquePaintEvent, false);
		setAttribute(Qt::WA_TranslucentBackground);
		setAutoFillBackground( false );
		setAttribute( Qt::WA_NoSystemBackground );

		ui->iconView->setAttribute(Qt::WA_TranslucentBackground);
		ui->iconView->setAutoFillBackground( false );

		this->layout()->addWidget( ui->iconView );

		ui->splitter->hide();
		connect(ui->iconView, SIGNAL(pressed(QModelIndex)), this, SIGNAL(mousePressed()));
	}
}

FileSystemWidget::~FileSystemWidget()
{
	delete ui;
}

void FileSystemWidget::initUi(QString openLocation)
{
	menu = new QMenu(this);
	noSelectionMenu = new QMenu(this);
	menu->setStyleSheet(STANDARDMENUSTYLE);
	noSelectionMenu->setStyleSheet(STANDARDMENUSTYLE);
	setMenu();
	setNoSelectionMenu();

	ui->placesView->setModel( static_cast<QStandardItemModel*>(placesModel) );
	ui->placesView->setIconSize( QSize( PLACES_ICON_SIZE, PLACES_ICON_SIZE ) );
	QFont placesFont;
	placesFont.setFamily( "Comic Sans MS" );
	ui->placesView->setFont( placesFont );

	initModel();

	ui->listView->setItemsExpandable( false );
	ui->listView->setIconSize( QSize( static_cast<int>(ICON_SIZE * 0.35), static_cast<int>(ICON_SIZE * 0.35) ));
	ui->listView->setStyleSheet( ui->listView->styleSheet() + " QTreeView::branch { border-image: url(nonexistent.png); }");
	ui->listView->setSortingEnabled( true );

	ui->iconView->setViewMode(QListView::IconMode);
	ui->iconView->setResizeMode( QListView::Adjust );
	ui->iconView->setMovement(QListView::Snap);
	if ( openLocation.isEmpty() )
	{
		if ( standalone )
		{
			setHome();
		} else {
			QString desktopDir = Settings::homePath() + QDir::separator() + "Desktop";
			if ( ! QFile::exists( desktopDir ) )
			{
				if ( ! QDir( Settings::homePath() ).mkdir( "Desktop" ) )
				{
					desktopDir = Settings::homePath();
				}
			}
			if ( QFile::exists( desktopDir ) )
			{
				setIndex( model->index( desktopDir ) );
			}
		}
	} else {
		QFileInfo openLocationInfo( openLocation );
		if ( openLocationInfo.exists() )
		{
			if ( openLocationInfo.isDir() )
			{
				setIndex( model->index( openLocationInfo.absoluteFilePath() ) );
			} else {
				setIndex( model->index( openLocationInfo.canonicalPath() ) );
			}
		} else {
			setHome();
		}
	}
	ui->listView->hide();

	QFont font;
	font.setFamily( Settings::getUserValue("Font", LIBREWORK_FONT).toString() );
	ui->addressLineEdit->setFont( font );
	ui->addressLineEdit->addAction(QIcon(ICON_DIR_32 + QString(QDir::separator()) + "folder-docs.png"), QLineEdit::LeadingPosition);
	ui->addressLineEdit->setReadOnly( true );

	QFont listViewFont;
	listViewFont.setFamily( "arial" ); //Settings::getUserValue("Font", LIBREWORK_FONT).toString() );
	listViewFont.setBold( false );
	//listViewFont.setPointSize( 12 );
	ui->listView->setFont( listViewFont );

	connect( ui->iconViewButton, SIGNAL(pressed()), ui->iconView, SLOT(show()));
	connect( ui->iconViewButton, SIGNAL(pressed()), ui->listView, SLOT(hide()));
	connect( ui->iconViewButton, SIGNAL(pressed()), this, SLOT(setEnabledButtons()));
	connect( ui->iconViewButton, SIGNAL(pressed()), this, SLOT(changeView()));

	connect( ui->listViewButton, SIGNAL(pressed()), ui->listView, SLOT(show()));
	connect( ui->listViewButton, SIGNAL(pressed()), ui->iconView, SLOT(hide()));
	connect( ui->listViewButton, SIGNAL(pressed()), this, SLOT(setEnabledButtons()));
	connect( ui->listViewButton, SIGNAL(pressed()), this, SLOT(changeView()));

	connect( ui->upButton, SIGNAL(pressed()), this, SLOT(upButtonPressed()));

	connect( ui->listView, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(doubleClicked(QModelIndex)));
	if ( ! standalone )
	{
		connect( ui->listView, SIGNAL(clicked(QModelIndex)), this, SLOT(doubleClicked(QModelIndex)));
	}
	connect( ui->listView, SIGNAL(pressed(QModelIndex)), this, SLOT(mousePressed(QModelIndex)));
	ui->listView->setContextMenuPolicy(Qt::CustomContextMenu);
	connect( ui->listView, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(menuRequested(QPoint)));
	connect( ui->listView, SIGNAL(dropFiles(QModelIndex,QStringList,QDropEvent *)), this, SLOT(dropFiles(QModelIndex,QStringList,QDropEvent *)));

	connect( ui->iconView, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(doubleClicked(QModelIndex)));
	if ( ! standalone )
	{
		connect( ui->iconView, SIGNAL(clicked(QModelIndex)), this, SLOT(doubleClicked(QModelIndex)));
	}
	connect( ui->iconView, SIGNAL(pressed(QModelIndex)), this, SLOT(mousePressed(QModelIndex)));
	ui->iconView->setContextMenuPolicy(Qt::CustomContextMenu);
	connect( ui->iconView, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(menuRequested(QPoint)));
	connect( ui->iconView, SIGNAL(dropFiles(QModelIndex,QStringList,QDropEvent *)), this, SLOT(dropFiles(QModelIndex,QStringList,QDropEvent *)));

	connect( ui->placesView, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(placesDoubleClicked(QModelIndex)));
	connect( ui->placesView, SIGNAL(clicked(QModelIndex)), this, SLOT(placesDoubleClicked(QModelIndex)));

	connect( ui->renameButton, SIGNAL(pressed()), this, SLOT(renameSlot()));
	connect( ui->newFolderButton, SIGNAL(pressed()), this, SLOT(newFolderSlot()));
	connect( ui->cutButton, SIGNAL(pressed()), this, SLOT(cutSlot()));
	connect( ui->copyButton, SIGNAL(pressed()), this, SLOT(copySlot()));
	connect( ui->pasteButton, SIGNAL(pressed()), this, SLOT(pasteSlot()));
	connect( ui->removeButton, SIGNAL(pressed()), this, SLOT(removeSlot()));
}

void FileSystemWidget::closeEvent(QCloseEvent * closeEvent)
{
	if ( standalone )
	{
		QStringList splitterGeometry;
		QList<int> sizes = ui->splitter->sizes();
		for ( int i = 0; i < sizes.count(); i++ )
		{
			splitterGeometry += QString::number( sizes.at(i) );
		}

		QStringList listViewGeometry;
		for ( int i = 0; i < 3; i++ )
		{
			listViewGeometry += QString::number( ui->listView->columnWidth( i ) );
		}

		QString viewType;
		if ( ui->listView->isVisible() )
		{
			viewType = "list";
		} else {
			viewType = "icon";
		}

		Settings::setUserValue("WindowGeometry/FileSystemWidget.splitter", QVariant(splitterGeometry));
		Settings::setUserValue("WindowGeometry/FileSystemWidget.listView", QVariant(listViewGeometry));
		Settings::setUserValue("WindowGeometry/FileSystemWidget.viewType", QVariant(viewType));
		Settings::saveWidgetGeometry("FileSystemWidget", this);
	} else {
		closeEvent->ignore();
	}
}

void FileSystemWidget::setHome()
{
	QModelIndex index = model->index( Settings::homePath() );
	setIndex( index );
}

void FileSystemWidget::directoryLoaded(QString dir)
{
	QString path = dir;
	updatePath( & path );
	if ( ! scrollTo.isEmpty() )
	{
		QModelIndex index = model->index(scrollTo);
		currentView->setCurrentIndex( index );
		currentView->scrollTo( model->index(scrollTo), QAbstractItemView::EnsureVisible );
		scrollTo.clear();
	}
	connectModelSignals();
	ui->fileInfo->setText(tr("Finished loading folder %1").arg(path));
}

void FileSystemWidget::fillPlaces()
{
	QFont font;
	font.setFamily( "arial" );
	placesModel = new PlacesModel(nullptr, true, font);
	updatePlaces();
}

QString FileSystemWidget::remotePlaceName(int i)
{
	QString place = tr("LOCAL > ") + remoteClientManager->getNames().at(i);
	QString drive = remoteClientManager->getDrives().at(i);
	if ( ! drive.isEmpty() )
	{
		place = place + " (" + drive.toUpper() + ")";
	}
	return place;
}

void FileSystemWidget::updatePlaces()
{
	placesModel->updatePlaces( remoteClientManager );
}

void FileSystemWidget::disconnectModelSignals()
{
    disconnect(model, SIGNAL(rootPathChanged(QString)), ui->listView->selectionModel(), SLOT(clearSelection()));
    disconnect(model, SIGNAL(rootPathChanged(QString)), ui->iconView->selectionModel(), SLOT(clearSelection()));
    disconnect( ui->iconView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), this, SLOT(setEnabledButtons()));
    disconnect( ui->listView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), this, SLOT(setEnabledButtons()));
}

void FileSystemWidget::connectModelSignals()
{
    connect(model, SIGNAL(rootPathChanged(QString)), ui->listView->selectionModel(), SLOT(clearSelection()));
    connect(model, SIGNAL(rootPathChanged(QString)), ui->iconView->selectionModel(), SLOT(clearSelection()));
    connect( ui->iconView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), this, SLOT(setEnabledButtons()));
    connect( ui->listView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), this, SLOT(setEnabledButtons()));
}

void FileSystemWidget::initModel()
{
	if ( model != nullptr)
    {
		disconnect(model, SIGNAL(directoryLoaded(QString)), this, SLOT(directoryLoaded(QString)));
		disconnectModelSignals();
    }

	FileSystemModel * newModel = new FileSystemModel( this );

	ui->listView->reset();
    ui->listView->setModel( newModel );
	ui->iconView->reset();
    ui->iconView->setModel( newModel );
	if ( model != nullptr)
    {
        delete model;
    }
    model = newModel;
    connect(model, SIGNAL(directoryLoaded(QString)), this, SLOT(directoryLoaded(QString)));
	connectModelSignals();
	setEnabledButtons();
}

void FileSystemWidget::placesDoubleClicked(QModelIndex index)
{
	if (index.isValid())
	{
		setPlaceIndex( index );
	}
}

void FileSystemWidget::setPlaceIndex( QModelIndex index )
{
	Place * place = placesModel->getItem( index );
	if ( place != nullptr)
    {
#ifdef Q_OS_WIN
		if ( place->location == qgetenv( "USERNAME" ) )
#else
		if ( place->location == qgetenv( "USER" ) )
#endif
		{
			setHome();
		} else {
			if ( place->location.startsWith('/') )
			{
				setIndex( model->index( place->location ) );
			} else {
				setIndex( model->index(Settings::homePath() + QDir::separator() + place->location ) );
			}
		}
    }
}

void FileSystemWidget::upButtonPressed()
{
	scrollTo = model->rootPath();
	QModelIndex index = model->index( scrollTo );
    bool allow = true;
	if ( QFileInfo(model->rootPath()).absoluteFilePath() == QFileInfo(Settings::homePath()).absoluteFilePath() )
    {
        allow = false;
    }
    for (int i = 0; i < remoteClientManager->getDrives().count(); i++)
    {
        QString path = model->rootPath();
        if ( path.right(1) == "/" )
        {
            path.chop(1);
        }
        QString drive = remoteClientManager->getDrives().at(i);
        if ( model->rootPath() == drive || path == drive )
        {
            allow = false;
        }
    }
    if (allow)
    {
        QString parentPath = model->filePath( model->parent( index ) );

		if ( parentPath != "/" )
		{
			setIndex( model->index( parentPath ) );
		}
    }
}

void FileSystemWidget::setIndex( QModelIndex index )
{
	if ( ! standalone )
	{
		if ( index != model->index(Settings::homePath() + QDir::separator() + "Desktop") )
		{
			return;
		}
	}
	if ( ! model->fileInfo( index ).exists() )
	{
		remoteClientManager->stop();
		remoteClientManager->start();
		updatePlaces();
		model->setRootPath( Settings::homePath() );
		return;
	}
    ui->listView->selectionModel()->clearSelection();
    ui->iconView->selectionModel()->clearSelection();
    disconnectModelSignals();
    QString path =  model->filePath( index );
    QString updatedPath = path;
    updatePath( &updatedPath );
    ui->fileInfo->setText( tr("Loading folder %1...").arg(updatedPath));
    model->setRootPath( path );
    ui->listView->setRootIndex( index );
    ui->iconView->setRootIndex( index );
    ui->addressLineEdit->setText( updatedPath );
}

void FileSystemWidget::updatePath(QString * path)
{
#ifdef Q_OS_WIN
	QString header = qgetenv( "USERNAME") + " > ";
#else
	QString header = qgetenv( "USER") + " > ";
#endif
	path->replace( QRegExp("^" + Settings::homePath()), header );
	for (int i = 0; i < remoteClientManager->getDrives().count(); i++)
	{
		path->replace( QRegExp("^" + remoteClientManager->getDrives().at(i)), tr("LOCAL > ") + remoteClientManager->getNames().at(i) );
	}
	if ( path->startsWith( header ) )
	{
		if ( path->mid( header.length(), 1) == QDir::separator() )
		{
			path->remove(header.length(), 1);
		}
	}
	path->replace( QDir::separator(), " > " );
	if ( path->right(1) == QDir::separator() )
	{
		path->chop(1);
	}
}

void FileSystemWidget::doubleClicked(QModelIndex index)
{
	if ( index.isValid() )
	{
		QFileInfo fileInfo = model->fileInfo( index );
		if ( fileInfo.isSymLink() )
		{
			fileInfo = QFileInfo( fileInfo.symLinkTarget() );
		}
		if ( ! fileInfo.isSymLink() )
		{
			if ( fileInfo.isDir() )
			{
				if ( standalone )
				{
					setIndex( index );
				} else {
					Application::launchFileManager( model->filePath( index ) );
					ui->iconView->clearSelection();
				}
			}
			if ( fileInfo.isFile() )
			{
				QString extension = fileInfo.suffix();
				if (Application::mimeType(fileInfo.absoluteFilePath()) == "recursivelink")
				{
					QMessageBox::critical ( nullptr, tr("Link error"), tr("This file is a recursive set of links"));
					return;
				}
				Application * preferredProgram = Application::preferredProgram(extension);
				if ( preferredProgram == nullptr )
				{
					openWithSlot();
					return;
				}
				Application * myApp = new Application( preferredProgram );
				myApp->setFile( model->filePath( index ) );
				myApp->launch();
				if ( ! standalone )
				{
					ui->iconView->clearSelection();
				}
			}
		} else {
			setHome();
		}
	}
}

void FileSystemWidget::mousePressed(QModelIndex index)
{
    // Right click
    if (QApplication::mouseButtons() == Qt::RightButton)
    {
		setEnabledButtons();
        showMenu( index );
    }
}

void FileSystemWidget::menuRequested(const QPoint pos)
{
	 setEnabledButtons();
    if (ui->listView->isVisible())
    {
        if (ui->listView->indexAt(pos) == QModelIndex())
        {
            showNoSelectionMenu();
        }
    } else {
        if (ui->iconView->indexAt(pos) == QModelIndex())
        {
            showNoSelectionMenu();
        }
    }
}

void FileSystemWidget::showMenu( QModelIndex index )
{
	openWithAction->setVisible( false );
	if (index.isValid())
	{
		bool multiple;
		bool onlyFiles = false;
		if ( model->fileInfo( index ).isFile() )
		{
			openWithAction->setVisible( true );
			onlyFiles = true;
		}

		if ( currentView == ui->iconView )
		{
			multiple = ( currentView->selectionModel()->selectedIndexes().count() > 1 );
		} else {
			multiple = ( currentView->selectionModel()->selectedRows().count() > 1 );
		}
		if ( multiple )
		{
			for ( int i = 0; i < currentView->selectionModel()->selectedIndexes().count(); i++ )
			{
				if ( model->fileInfo( currentView->selectionModel()->selectedIndexes().at(i) ).isDir() )
				{
					onlyFiles = false;
				}
			}
		}
		openRemotelyAction->setEnabled( onlyFiles );
		openRemotelyAction->setVisible( true );
		openWithAction->setEnabled( ! multiple );
		renameAction->setEnabled( ! multiple );
		propertiesAction->setEnabled( ! multiple );
	}
    menu->exec(QCursor::pos());
}

void FileSystemWidget::showNoSelectionMenu()
{
    noSelectionMenu->exec(QCursor::pos());
}

void FileSystemWidget::close()
{
	if ( model != nullptr )
	{
		delete model;
	}

	if ( remoteClientManager != nullptr )
	{
		remoteClientManager->stop();
		delete remoteClientManager;
	}
    QWidget::close();
}

void FileSystemWidget::setMenu()
{
    renameAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "rename.png"), tr("&Rename"), this );
    menu->addAction( renameAction );
    connect( renameAction, SIGNAL(triggered()), this, SLOT(renameSlot()));

    cutAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "cut.png"), tr("Cu&t"), this );
    menu->addAction( cutAction );
    connect( cutAction, SIGNAL(triggered()), this, SLOT(cutSlot()));

    copyAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "copy.png"), tr("&Copy"), this );
    menu->addAction( copyAction );
    connect( copyAction, SIGNAL(triggered()), this, SLOT(copySlot()));

    removeAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "remove.png"), tr("&Delete"), this );
    menu->addAction( removeAction );
    connect( removeAction, SIGNAL(triggered()), this, SLOT(removeSlot()));

	openWithAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "openwith.png"), tr("&Open with..."), this );
	menu->addAction( openWithAction );
	connect( openWithAction, SIGNAL(triggered()), this, SLOT(openWithSlot()));

	openRemotelyAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "forward.png"), tr("&Open remotely"), this );
	menu->addAction( openRemotelyAction );
	connect( openRemotelyAction, SIGNAL(triggered()), this, SLOT(openRemotelySlot()));

	propertiesAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "openwith.png"), tr("&Properties"), this );
	menu->addAction( propertiesAction );
	connect( propertiesAction, SIGNAL(triggered()), this, SLOT(filePropertiesSlot()));
}

void FileSystemWidget::setNoSelectionMenu()
{
	newFolderAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "new-folder.png"), tr("&New folder"), this );
	noSelectionMenu->addAction( newFolderAction );
	connect( newFolderAction, SIGNAL(triggered()), this, SLOT(newFolderSlot()));

	pasteAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "paste.png"), tr("&Paste"), this );
	noSelectionMenu->addAction( pasteAction );
	connect( pasteAction, SIGNAL( triggered()), this, SLOT(pasteSlot()));

	findFileAction = new QAction( QIcon(ICON_DIR_32 + QString(QDir::separator()) + "search.png"), tr("&Find file"), this );
	noSelectionMenu->addAction( findFileAction );
	connect( findFileAction, SIGNAL( triggered()), this, SLOT(findFileSlot()));
}

void FileSystemWidget::dropFiles(QModelIndex target, QStringList fileList, QDropEvent *e)
{
	if ( target.isValid() )
	{
		QList<QAction*> actionList;
		actionList += new QAction(QIcon(":/icons/32x32/cut.png"), tr("Move files"));
		actionList += new QAction(QIcon(":/icons/32x32/copy.png"), tr("Copy files"));
		//actionList += new QAction(QIcon(":/icons/32x32/link.png"), tr("Create shortcut"));
		QAction * selectedAction = QMenu::exec( actionList, QCursor::pos() );
		if ( selectedAction == nullptr )
		{
			return;
		}

		QString targetDir;
		QFileInfo targetFileInfo = model->fileInfo( target );
		if ( targetFileInfo.isDir() )
		{
			targetDir = targetFileInfo.absoluteFilePath();
		} else {
			targetDir = targetFileInfo.absoluteDir().absolutePath();
		}
		QModelIndexList indexes;
		for ( int i = 0; i < fileList.count(); i++ )
		{
			QFileInfo fileListInfo = QFileInfo( fileList[i].trimmed() );
			if ( fileListInfo.absoluteDir().absolutePath() == targetDir )
			{
				QMessageBox::critical(nullptr, tr("Error droping files"), tr("Cannot drop file %1 into the same folder").arg(fileList[i].trimmed()));
				return;
			}
			QModelIndex index = model->index( fileList[i].trimmed() );
			if ( ! index.isValid() )
			{
				QMessageBox::critical(nullptr, tr("Error droping files"), tr("Cannot drop file %1").arg(fileList[i].trimmed()));
				return;
			} else {
				indexes += index;
			}
		}

		if ( selectedAction == actionList[0] ) // Move files
		{
			model->moveFiles( indexes, QFileInfo( targetDir ) );
		}
		if ( selectedAction == actionList[1] ) // Copy files
		{
			model->copyFiles( indexes, QFileInfo( targetDir ) );
		}
		if ( targetFileInfo.isDir() )
		{
			setIndex( target );
		}
	}
	e->acceptProposedAction();
}

void FileSystemWidget::cutSlot()
{
	QList<QFileInfo> copiedFiles;
	int count = currentView->selectionModel()->selectedIndexes().count();
	for ( int i = 0; i < count; i++ )
	{
		copiedFiles += model->fileInfo( currentView->selectionModel()->selectedIndexes().at(i) );
	}
	ClipboardFiles::setEntries( copiedFiles, cut );
	if ( count > 0 )
	{
		cut = true;
	}
	setEnabledButtons();
}

void FileSystemWidget::copySlot()
{
	QList<QFileInfo> copiedFiles;
	int count = currentView->selectionModel()->selectedIndexes().count();
	for ( int i = 0; i < count; i++ )
	{
		copiedFiles += model->fileInfo( currentView->selectionModel()->selectedIndexes().at(i) );
	}
	ClipboardFiles::setEntries( copiedFiles, cut );
	cut = false;
	setEnabledButtons();
}

void FileSystemWidget::pasteSlot()
{
	QModelIndexList indexList;
	QString exists;
	QPair<QList<QFileInfo>,bool> entries = ClipboardFiles::entries();
	QList<QFileInfo> copiedFiles = entries.first;
	cut = entries.second;

	for ( int i = 0; i < copiedFiles.count(); i++ )
	{
		qDebug("Pasting file '%s'",  copiedFiles.at(i).absoluteFilePath().trimmed().toUtf8().data());
		if ( copiedFiles.at(i).exists() )
		{
			exists += copiedFiles.at(i).fileName() + "\n";
		} else {
			selectedIndex = model->index(copiedFiles.at(i).absoluteFilePath().trimmed());
			if ( selectedIndex.isValid() )
			{
				indexList += selectedIndex;
			} else {
				exists += copiedFiles.at(i).fileName() + "\n";
			}
		}
	}
	if ( ! exists.isEmpty() )
	{
		QMessageBox::critical( this, tr("Cannot paste"), tr("The target files you have selected currently exist:\n%1").arg( exists ), QMessageBox::Ok, QMessageBox::NoButton);
		return;
	} else {
		if ( cut )
		{
			model->moveFiles( indexList, QFileInfo(model->rootPath()) );
		} else {
			model->copyFiles( indexList, QFileInfo(model->rootPath()) );
		}
	}
	cut = false;
	selectedIndex = QModelIndex();
	setEnabledButtons();
}

void FileSystemWidget::removeSlot()
{
    bool doIt = true;

    QString title;
    QString message;
	QModelIndexList selection = currentView->selectionModel()->selectedIndexes();
	bool multiple = ( selection.count() >= 1 );
	if ( selection.count() == 0 )
	{
		return;
	}
	if ( multiple )
	{
		title = tr("Remove selection");
		message = tr("All selected files and folders are to be removed with all its contents. This operation cannot be undone. Are you sure you want to remove selected folders and files?");
	} else {
		if ( model->fileInfo( selectedIndex ).isDir() )
		{
			title = tr("Remove folder");
			message = tr("This folder is to be removed with all its contents. This operation cannot be undone. Are you sure you want to remove this folder?");
		} else {
			title = tr("Remove file");
			message = tr("This file is to be removed. This operation cannot be undone. Are you sure you want to remove this file?");
		}
	}

	QMessageBox * removeMessageBox = new QMessageBox( QMessageBox::Warning, title, message, QMessageBox::Yes | QMessageBox::No, this );
	removeMessageBox->setDefaultButton(  QMessageBox::No );
	removeMessageBox->setStyleSheet(QString( STANDARDDIALOGSTYLE ));
	removeMessageBox->button( QMessageBox::Yes )->setStyleSheet(QString( STANDARDDIALOGSTYLE ));
	removeMessageBox->button( QMessageBox::No )->setStyleSheet(QString( STANDARDDIALOGSTYLE ));
	doIt = removeMessageBox->exec() == QMessageBox::Yes;
	delete removeMessageBox;

	if (doIt)
	{
		model->removeFiles( selection );
		selectedIndex = QModelIndex();
		ui->listView->selectionModel()->clearSelection();
		ui->iconView->selectionModel()->clearSelection();
	}
}

void FileSystemWidget::openFile()
{
	QAction * _sender = static_cast<QAction*>(sender());
	QStringList data = _sender->data().toStringList();
	if (data.count() == 2)
	{
		Application * app = new Application( reinterpret_cast<Application*>(data[1].toLongLong()) );
		app->setFile( data[0] );
		app->launch();
	}
}

void FileSystemWidget::openRemotelySlot()
{
	QStringList fileList;
	QString serverAddress = Settings::getValue( "Server/LibreworkServerAddress" );
	int serverPort = Settings::getValue( "Server/LibreworkServerPort" ).toInt();
	QString targetDir = QUuid::createUuid().toString();

	for ( int i = 0; i < currentView->selectionModel()->selectedIndexes().count(); i ++ )
	{
		bool onlyCopyFile = true;
		if ( i == currentView->selectionModel()->selectedIndexes().count() - 1 ) // last Item
		{
			onlyCopyFile = false;
		}
		QFileInfo fileInfo = model->fileInfo(currentView->selectionModel()->selectedIndexes().at(i));
		if ( ! fileList.contains(fileInfo.absoluteFilePath()) )
		{
			fileList += fileInfo.absoluteFilePath();
		}
	}

	qDebug("Filelist: '%s'", fileList.join(',').toUtf8().data());
	FileThread * fileThread = new FileThread( FileThread::REMOTE_COPY, nullptr, fileList, targetDir );
	FileThreadWidget * fileThreadWidget = new FileThreadWidget( fileThread, nullptr );
	connect( fileThread, SIGNAL(finished(QStringList,QString)), this, SLOT(openRemotelySlotFinished(QStringList,QString)));
	fileThreadWidget->show();
	fileThreadWidget->start();
}

void FileSystemWidget::openRemotelySlotFinished(QStringList fileNames, QString targetDir)
{
	qDebug("Filenames: '%s'", fileNames.join(',').toUtf8().data());
	if ( fileNames.count() > 1 )
	{
		Manager::openRemoteServerDir( targetDir );
	} else {
		Manager::launchRemoteApplication( QFileInfo(fileNames[0]).fileName(), targetDir );
	}
}

void FileSystemWidget::openWithSlot()
{
	QList<Application*> applications;
	QFileInfo file = model->fileInfo( selectedIndex );
	if ( file.isFile() )
	{
		QMenu * openWithMenu = new QMenu(this);
		openWithMenu->setStyleSheet(STANDARDMENUSTYLE);
		openWithMenu->setTitle(tr("Open with..."));
		openWithMenu->addAction(QIcon(ICON_DIR_32 + QString(QDir::separator()) + "openwith.png"), tr("Open with..."));
		openWithMenu->addSeparator();
		QString extension = QFileInfo( file ).suffix();
		if (Application::mimeType(file) == "recursivelink")
		{
			QMessageBox::critical ( nullptr, tr("Link error"), tr("This file is a recursive set of links"));
			return;
		}
		applications += Application::applicationList( extension );
		int foundByExtension = applications.count();
		applications += Application::applicationList();

		for ( int i = 0; i < applications.count(); i++ )
		{
			if ( i == foundByExtension )
			{
				openWithMenu->addSeparator();
			}
			QIcon icon = applications[i]->getIcon();
			QString name = applications[i]->getName();
			QAction * action = openWithMenu->addAction( icon, name, this, SLOT(openFile()));
			QStringList data;
			data += file.absoluteFilePath();
			data += QString::number( reinterpret_cast<long long>(applications[i]) );
			action->setData( data );
		}
		openWithMenu->exec(QCursor::pos());
		connect(openWithMenu, SIGNAL(aboutToHide()), openWithMenu, SLOT(deleteLater()));
	}
}

void FileSystemWidget::newFolderSlot()
{
    QString newFolderName = QInputDialog::getText( this, tr("New folder"), tr("Please, enter new folder name"));
    if ( ! newFolderName.isEmpty() )
    {
        if ( QFile::exists( model->rootPath() + QDir::separator() + newFolderName))
        {
            QMessageBox::critical( this, tr("Cannot create folder"), tr("The folder name you have selected currently exists. Please, select a different folder name"), QMessageBox::Ok, QMessageBox::NoButton);
        } else {
            model->mkdir( model->index( model->rootPath()), newFolderName );
        }
    }
}

void FileSystemWidget::renameSlot()
{
    if ( model->fileInfo( selectedIndex ).isDir() )
    {
		QString newName = QInputDialog::getText( nullptr, tr("Rename"), tr("Please, enter new folder name"), QLineEdit::Normal, model->fileName( selectedIndex ) );
        if ( newName !=  model->fileName( selectedIndex ) && ! newName.isEmpty() )
        {
            if ( QFile::exists( model->rootPath() + QDir::separator() + newName))
            {
                QMessageBox::critical( this, tr("Cannot rename folder"), tr("The folder name you have selected currently exists. Please, select a different folder name"), QMessageBox::Ok, QMessageBox::NoButton);
            } else {
				model->renameFile( selectedIndex, newName );
            }
        }
    } else {
		QString newName = QInputDialog::getText( nullptr, tr("Rename"), tr("Please, enter new file name"), QLineEdit::Normal, model->fileName( selectedIndex ) );
        if ( newName !=  model->fileName( selectedIndex ) && ! newName.isEmpty() )
        {
            if ( QFile::exists( model->rootPath() + QDir::separator() + newName))
            {
                QMessageBox::critical( this, tr("Cannot rename file"), tr("The file name you have selected currently exists. Please, select a different file name"), QMessageBox::Ok, QMessageBox::NoButton);
            } else {
				model->renameFile( selectedIndex, newName );
            }
        }
    }
}

void FileSystemWidget::filePropertiesSlot()
{
	FileProperties * properties = new FileProperties( model->fileInfo(selectedIndex) );
	properties->show();
	connect( properties, SIGNAL(accepted()), properties, SLOT(deleteLater()));
	connect( properties, SIGNAL(rejected()), properties, SLOT(deleteLater()));
}

void FileSystemWidget::setEnabledButtons()
{
        if (ClipboardFiles::isEmpty())
	{
		pasteAction->setEnabled( false );
		ui->pasteButton->setEnabled( false );
	} else {
		pasteAction->setEnabled( true );
		ui->pasteButton->setEnabled( true );
	}
    if (ui->listView->isVisible())
    {
		currentView = ui->listView;
        ui->listViewButton->setEnabled( false );
        ui->iconViewButton->setEnabled( true );
    } else {
		currentView = ui->iconView;
        ui->listViewButton->setEnabled( true );
        ui->iconViewButton->setEnabled( false );
    }
    if ( ui->listView->selectionModel()->hasSelection() || ui->iconView->selectionModel()->hasSelection() )
    {
        if ( ui->listView->selectionModel()->hasSelection() )
        {
            selectedIndex = ui->listView->selectionModel()->selectedIndexes()[0];
        } else {
            selectedIndex = ui->iconView->selectionModel()->selectedIndexes()[0];
        }
        renameAction->setEnabled( true );
        ui->renameButton->setEnabled( true );
        copyAction->setEnabled( true );
        ui->copyButton->setEnabled( true );
        cutAction->setEnabled( true );
        ui->cutButton->setEnabled( true );
        removeAction->setEnabled( true );
        ui->removeButton->setEnabled( true );
    } else {
        selectedIndex = QModelIndex();
        renameAction->setEnabled( false );
        ui->renameButton->setEnabled( false );
        copyAction->setEnabled( false );
        ui->copyButton->setEnabled( false );
        cutAction->setEnabled( false );
        ui->cutButton->setEnabled( false );
        removeAction->setEnabled( false );
        ui->removeButton->setEnabled( false );
    }
}

void FileSystemWidget::changeView()
{
    if (  ui->listView->selectionModel()->hasSelection() )
    {
        ui->iconView->selectionModel()->select( ui->listView->selectionModel()->selection(), QItemSelectionModel::ClearAndSelect );
    } else {
        ui->listView->selectionModel()->select( ui->iconView->selectionModel()->selection(), QItemSelectionModel::ClearAndSelect );
    }
}

void FileSystemWidget::findFileSlot()
{
	bool alreadyFound = false;
	QString ret;
	QString searchString = QInputDialog::getText( this, tr("Find file"), tr("Please, enter the search string"), QLineEdit::Normal );
	if ( ! searchString.isEmpty() )
	{
		bool cancelSearch = false;
		mutex.lock();
		searching = true;
		mutex.unlock();
		if ( searchProgress != nullptr )
		{
			delete searchProgress;
		}
		searchProgress = new QProgressDialog(QString(), tr("Cancel search"), 0, 0, this);
		QLabel * searchLabel = new QLabel(tr("Search progress..."));
		searchLabel->setFixedWidth(300);
		searchLabel->setAlignment(Qt::AlignLeft | Qt::AlignVCenter);
		searchProgress->setLabel( searchLabel );
		searchProgress->setWindowTitle(tr("Find file"));
		searchProgress->setWindowIcon(QIcon(":/icons/32x32/search.png"));
		connect(searchProgress, SIGNAL(canceled()), this, SLOT(cancelSearch()));
		searchProgress->show();
		qApp->processEvents();

		QDirIterator it(model->rootPath(), QDirIterator::Subdirectories);
		while (it.hasNext() && ! cancelSearch)
		{
			it.next();
			searchProgress->setLabelText(tr("File name: %1").arg(it.fileName()));
			if (it.fileName().contains( searchString, Qt::CaseInsensitive) )
			{
				ret = it.filePath();
				if ( ! ret.isEmpty() )
				{
					QModelIndex index = model->index(ret);
					if ( index.isValid() )
					{
						alreadyFound = true;
						if (model->fileInfo(index).isDir())
						{
							setIndex(index);
						} else {
							QModelIndex parent = index.parent();
							setIndex(parent);
							if (ui->iconView->isVisible())
							{
								ui->iconView->scrollTo( index );
							}
							if (ui->listView->isVisible())
							{
								ui->listView->scrollTo( index );
							}
							searchProgress->hide();
							if ( QMessageBox::information( this,
												   tr("Find file"),
												   tr("Found file '%1'. Do you want to continue searching?").arg( ret ),
												   QMessageBox::Yes,
												   QMessageBox::No) == QMessageBox::No )
							{
								searchProgress->cancel();
								delete searchProgress;
								searchProgress = nullptr;
								return;
							} else {
								searchProgress->show();
							}
						}
					}
				}
			}
			mutex.lock();
			cancelSearch = ! searching;
			mutex.unlock();
			qApp->processEvents();
		}
	} else {
		return;
	}

	if (searchProgress != nullptr)
	{
		searchProgress->cancel();
		delete searchProgress;
		searchProgress = nullptr;
	}

	if ( ! alreadyFound )
	{
		QMessageBox::warning( this, tr("Find file"), tr("No file or folder found matching your search"));
	} else {
		QMessageBox::information( this, tr("Find file"), tr("No more files or folders found"));
	}
}

void FileSystemWidget::cancelSearch()
{
	mutex.lock();
	searching = false;
	mutex.unlock();
}

void FileSystemWidget::keyPressEvent(QKeyEvent *event)
{
	if ( event->type() == QKeyEvent::KeyPress )
	{
		if ( event->matches(QKeySequence::Copy) )
		{
			copySlot();
		}
		if ( event->matches(QKeySequence::Cut) )
		{
			cutSlot();
		}
		if ( event->matches(QKeySequence::Paste) )
		{
			pasteSlot();
		}
		if ( event->matches(QKeySequence::Delete) )
		{
			removeSlot();
		}
	}
}
