#include "libreworkdesktopinterface.h"

#include <QApplication>
#include <QStyle>
#include <QStyleFactory>
#include <QTranslator>
#include <QProcess>
#include <QFontDatabase>
#include <QFont>
#include <QDir>
#include <QUuid>

#ifdef Q_OS_LINUX
#include <unistd.h>
#include <sys/types.h>
#include <librework-client/clientsession.h>
#endif

#include "initprogress.h"
#include "manager.h"
#include "constants.h"
#include "settings.h"
#include "lwlocale.h"
#include "remoteappserver.h"

#include <iostream>

#ifdef Q_OS_WIN
#include <windows.h>
#include <fcntl.h>
#include <io.h>
#else
#include <termios.h>
#include <unistd.h>
#endif

void usage(QString command)
{
	qDebug("Usage: %s [-i | -a <authFile>] [ -w ] [ -f [ <url to open> ] | -r <remote desktop connection name> ]", command.toUtf8().data());
	qDebug("Options:");
	qDebug("  (by default: launch desktop)");
	qDebug("  -f   Launch File Manager");
	qDebug("  -w   Launch Remote Application Server");
#ifdef Q_OS_LINUX
	qDebug("  -r   Launch Remote Desktop Connection");
#endif
	qDebug("  -a   Get credentials from file");
	qDebug("       With File Manger reads <username>, <password>, <server address>  and <server port> line by line from file");
#ifdef Q_OS_LINUX
	qDebug("       With Remote Desktop Connection reads <username>, <password> and ( <server config> or <remote server:port> ) line by line from file");
#endif
	qDebug("  -i   Get credentials interactively");
	qDebug("       With File Manger reads <username>, <password>, <server address> and <server port> from stdin");
#ifdef Q_OS_LINUX
	qDebug("       With Remote Desktop Connection reads <username>, <password> and ( <serverconfig> or <remote server:port> ) from stdin");
#endif
}

void setStdinEcho(bool enable = true)
{
#ifdef Q_OS_WIN
	HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
	DWORD mode;
	GetConsoleMode(hStdin, &mode);

	if( ! enable )
	{
		mode &= ~ENABLE_ECHO_INPUT;
	} else {
		mode |= ENABLE_ECHO_INPUT;
	}

	SetConsoleMode(hStdin, mode );
#else
	struct termios tty;
	tcgetattr(STDIN_FILENO, &tty);
	if( !enable )
	{
		tty.c_lflag &= ~ECHO;
	} else {
		tty.c_lflag |= ECHO;
	}
	(void) tcsetattr(STDIN_FILENO, TCSANOW, &tty);
#endif
}

int runLibreworkDesktop(int argc, char **argv)
{
	QApplication qApplication(argc, argv);
	qRegisterMetaType<HWND>("HWND");

#ifdef Q_OS_WIN
	QString username = qgetenv("USERNAME");
#else
	QString username = qgetenv("USER");
#endif
	QString password;
	QString serverAddress = Settings::getServerAddress();
	QString serverConfig;
	int serverPort = Settings::getServerPort();

	QString kioskApp = Settings::getKioskApp();

	Settings::init();
	Settings::setUserName( username );

	QDir fontsDir( LIBREWORK_FONT_DIR );
	QStringList fontFile = fontsDir.entryList(QStringList("*.ttf"));
	QList<int> fontId;
	for (int i = 0; i < fontFile.count(); i++)
	{
		fontId.append( QFontDatabase::addApplicationFont( LIBREWORK_FONT_DIR + QString(QDir::separator()) + fontFile.at(i) ) );
	}

	if ( fontId.count() > 0 )
	{
		QFontDatabase fontDatabase;
		QString family = QFontDatabase::applicationFontFamilies(fontId.at(0)).at(0);
		QFont font = fontDatabase.font( Settings::getUserValue("Font", LIBREWORK_FONT).toString(), QString(), Settings::getUserValue("FontSize", QString::number(LIBREWORK_FONT_BASE_SIZE)).toInt());
		qApplication.setFont( font );
	}

	QStyle *style = QStyleFactory::create("plastique");

	if(style)
	{
		QApplication::setStyle(style);
	}

	Settings::setExtensions();

#ifdef Q_OS_LINUX
	QFile layoutConfFile( "/etc/default/keyboard" );
	bool found = false;
	QByteArray line;

	if ( layoutConfFile.open( QIODevice::ReadOnly ) )
	{
		while ( ! found )
		{
			line = layoutConfFile.readLine();
			QString start = "XKBLAYOUT=";
			if ( line.trimmed().startsWith( start.toUtf8() ) )
			{
				found = true;
			}
		}
	}

	QString layout = "us";

	if ( found )
	{
		layout =  QString( line ).section( '=', 1, 1 );
		layout.remove( '"' );
		layout = layout.simplified();
	}
	QString program("setxkbmap");
	QStringList arguments;
	arguments += "-layout";
	arguments += Settings::getUserValue( "KeyboardLayout", layout ).toString();

	QProcess::execute( program, arguments );
#endif

	QTranslator translator;

	if (translator.load("librework-desktop_" + Settings::getUserValue("Language", "es").toString()))
	{
		qApplication.installTranslator(&translator);
	}

#ifdef Q_OS_LINUX
	QFile confFile( "/etc/default/locale" );
	found = false;

	if ( confFile.open( QIODevice::ReadOnly ) )
	{
		while ( ! found )
		{
			line = confFile.readLine();
			QString start = "LANG=";
			if ( line.trimmed().startsWith( start.toUtf8() ) )
			{
				found = true;
			}
		}
	}

	QString lang = "en_US.UTF-8";

	if ( found )
	{
		lang =  QString( line ).section( '=', 1, 1 );
		lang.remove( '"' );
		lang = lang.simplified();
	}

	QString language = Settings::getUserValue("Language", "").toString();
	if ( ! language.isEmpty() )
	{
		lang = LWLocale::getLang( language );
	}
	qputenv("LANG", lang.toUtf8());
#endif

	int ret;

	QStringList appArguments = QCoreApplication::arguments();

	int numberOfStdinArguments = 0;
	bool fileManager = false;
	bool remoteApplicationServer = false;
	//bool launchRemoteApplication = false;
	bool remoteDesktop = false;
	bool readFromStdin = false;
	bool readFromFile = false;
	QString authFileString;
	QString openLocation;
	QStringList remoteApplicationArguments;

	int pos = appArguments.indexOf("-i", 0); // Read password from stdin option
	if ( pos != -1 )
	{
		numberOfStdinArguments = 4; // By default, if launching desktop. Otherwise it will change later...
		readFromStdin = true;
		appArguments.removeAt( pos );
	}

	pos = appArguments.indexOf("-w", 0); // Launch remote application server
	if ( pos != -1 )
	{
		remoteApplicationServer = true;
		appArguments.removeAt( pos );
	}

	pos = appArguments.indexOf("-a", 0); // Read password from file option
	if ( pos != -1 )
	{
		if ( appArguments.count() > pos + 1 )
		{
			authFileString = appArguments[pos + 1];
			readFromFile = true;
			if ( readFromStdin )
			{
				qDebug("readFromStdin and readFromFile cannot be together");
				usage( appArguments[0] );
				exit(1);
			}
			appArguments.removeAt( pos + 1 );
			appArguments.removeAt( pos );
		} else {
			usage( appArguments[0] );
			exit(1);
		}
	}

	pos = appArguments.indexOf("-f", 0); // Launch file manager instead of desktop
	if ( pos != -1 )
	{
		fileManager = true;
		numberOfStdinArguments = 4;
		if ( appArguments.count() > pos + 1 )
		{
			openLocation = appArguments[pos + 1];
			appArguments.removeAt( pos + 1 );
		}
		appArguments.removeAt( pos );
	}
#ifdef Q_OS_LINUX
	pos = appArguments.indexOf("-r", 0); // Launch remote desktop connection instead of desktop
	if ( pos != -1 )
	{
		numberOfStdinArguments = 3;
		if ( fileManager )
		{
			usage( appArguments[0] );
			return -1;
		}
		remoteDesktop = true;
		appArguments.removeAt( pos );
	}
#endif
	/*pos = appArguments.indexOf("-t", 0); // Launch remote application instead of desktop
	if ( pos != -1 )
	{
		if ( fileManager || remoteDesktop )
		{
			usage( appArguments[0] );
			return -1;
		}
		launchRemoteApplication = true;
		while ( appArguments.count() > pos + 1 )
		{
			remoteApplicationArguments += appArguments[pos + 1];
			appArguments.removeAt( pos + 1 );
		}
		appArguments.removeAt( pos );
	}*/
	pos = appArguments.indexOf("--help", 0); // Launch remote desktop connection instead of desktop
	if ( pos != -1 )
	{
		usage( appArguments[0] );
		return -1;
	}

	if ( appArguments.count() > 1 )
	{
		openLocation = appArguments[1];
		appArguments.removeAt( 1 );
	}

	if ( appArguments.count() > 1 )
	{
		usage( appArguments[0] );
		return -1;
	}

	if ( readFromStdin )
	{
#ifdef Q_OS_WIN32
		_setmode(_fileno(stdin), _O_BINARY);
#endif
		qDebug("Waiting username,password,serveraddress,serverport from stdin...");
		setStdinEcho( false );

		QStringList stdinData;
		for ( int i = 0; i < numberOfStdinArguments; i++ )
		{
			QString string;
			bool exitLoop = false;
			while ( ! exitLoop )
			{
				char arr[1];
				std::cin.read(arr,sizeof(arr));
				if ( arr[0] == '\n' )
				{
					exitLoop = true;
				} else {
					string += arr[0];
				}
			}
			stdinData.append( string );
		}
		if ( stdinData.count() == 3 )
		{
			username = stdinData[0];
			password = stdinData[1];
			serverConfig = stdinData[2];
		}
		if ( stdinData.count() == 4 )
		{
			username = stdinData[0];
			password = stdinData[1];
			serverAddress = stdinData[2];
			serverPort = stdinData[3].toInt();
		}

		setStdinEcho( true );
	}

	if ( readFromFile )
	{
		int numLines = 4; // By default, if launching local Desktop
		if ( remoteDesktop )
		{
			numLines = 3;
		}
		if ( fileManager )
		{
			numLines = 4;
		}
		QStringList lines;
		QFile authFile( authFileString );
		authFile.open( QFile::ReadOnly );
		for ( int i = 0; i < numLines; i++ )
		{
			lines += authFile.readLine();
			if ( lines[i].size() < 0 )
			{
				authFile.close();
				authFile.remove();
				qDebug("Malformed auth file");
				usage( appArguments[0] );
				exit(1);
			}
		}
		authFile.close();
		authFile.remove();
		if ( numLines == 3 )
		{
			username = lines[0].simplified();
			password = lines[1].simplified();
			serverConfig = lines[2].simplified();
		}
		if ( numLines == 4 )
		{
			username = lines[0].simplified();
			password = lines[1].simplified();
			serverAddress = lines[2].simplified();
			serverPort = lines[3].simplified().toInt();
		}
	}

	/*if ( launchRemoteApplication )
	{
		if ( remoteApplicationArguments.count() > 0 )
		{
			serverPort = Settings::getValue( "Server/LibreworkServerPort" ).toInt();
			serverAddress = Settings::getValue( "Server/LibreworkServerAddress" );
			QString fileName = remoteApplicationArguments[0];
			QString targetDir;
			QString remoteApplication;
			if ( remoteApplicationArguments.count() > 1 )
			{
				targetDir = remoteApplicationArguments[1];
			}
			if ( remoteApplicationArguments.count() > 2 )
			{
				targetDir = remoteApplicationArguments[2];
			}
			if ( targetDir.isEmpty() )
			{
				targetDir = QUuid::createUuid().toString();
			}
			Manager::copyFilesToProfilesServer( QStringList( fileName), targetDir );
			Manager::launchRemoteApplication( fileName, targetDir, remoteApplication );
			return 0;
		} else {
			qDebug("-t option must have at least an argument (the file name to open remotely)");
			usage( appArguments[0] );
			return -1;
		}
	}*/

	if ( fileManager )
	{
		Settings::setUserName( username );
		Settings::setPassword( password );
		Settings::setServerAddress( serverAddress );
		Settings::setServerPort( serverPort );

		if (  serverPort > 0 )
		{
			QString token;
			if ( password.isEmpty() )
			{
				qDebug("Looking for token on file: %s", QString(TOKEN_FILE).toUtf8().data());
				if (QFile::exists( TOKEN_FILE ))
				{
					token = Settings::getValue( TOKEN_FILE, "UserToken" );
				}
			} else {
				token = password;
			}

			if ( ! password.isEmpty() )
			{
				if ( LWClient::instance()->connect() )
				{
					if ( LWClient::instance()->authenticate( Settings::getFullUserName(), token ) )
					{
						qDebug("Succesfully authenticated user token on Librework Server");
					} else {
						qDebug("Cannot authenticate on Librework server with user %s and token %s", Settings::getFullUserName().toUtf8().data(), token.toUtf8().data());
					}
				}
			}
		}

		Application::updateApplicationCache();

		Settings::setExtensions();

		FileSystemWidget * fileSystemWidget = new FileSystemWidget(nullptr, true, openLocation);

		fileSystemWidget->show();
		ret = qApplication.exec();
		delete fileSystemWidget;
	} else if (remoteApplicationServer) {
		RemoteAppServer * server = new RemoteAppServer();
		ret = server->startServer();
		delete server;
	} else if ( ! kioskApp.isEmpty() ) {
#ifdef Q_OS_LINUX
		QProcess::startDetached("amixer set Master 100%");
		Manager::startWindowManager();
#endif
		QStringList kioskAppList = kioskApp.split(" ");
		QString kioskAppCommand = kioskAppList[0];
		kioskAppList.removeAt( 0 );
		ret = QProcess::execute( kioskAppCommand, kioskAppList );
	} else {
#ifdef Q_OS_LINUX
		QProcess::startDetached("amixer set Master 100%");

		if ( remoteDesktop )
		{
			Manager::generateMimeAppsList();
			Manager::startWindowManager( true );
			ClientSession * clientSession = new ClientSession( 0, 0, true );
			clientSession->launchRemovableManagerProcess();
                        clientSession->launchProcess(username, password, serverConfig );
			qDebug("Username: %s, password: %s, serverconfig: %s", username.toUtf8().data(), password.toUtf8().data(), serverConfig.toUtf8().data());
			QTimer::singleShot(1000, clientSession, SLOT(showContainer()));
			QObject::connect( clientSession, SIGNAL(finished()), qApp, SLOT(quit()));
			ret = qApplication.exec();
			delete clientSession;
			Manager::stopWindowManager();
		} else {
#endif
			InitProgress * initProgress = new InitProgress(); // class Desktop will take care of its deletion
			initProgress->setGeometry( ( qApp->primaryScreen()->geometry().width() - INITPROGRESS_WIDTH ) / 2, ( qApp->primaryScreen()->geometry().height() - INITPROGRESS_HEIGHT ) / 2, INITPROGRESS_WIDTH, INITPROGRESS_HEIGHT );
			initProgress->show();
			if ( serverAddress.isEmpty() && ! readFromStdin )
			{
				serverPort = Settings::getValue( "Server/LibreworkServerPort" ).toInt();
				serverAddress = Settings::getValue( "Server/LibreworkServerAddress" );
			}
			Manager * manager = new Manager(initProgress, username, password, serverAddress, serverPort);
			QWebEngineProfile::defaultProfile()->setHttpAcceptLanguage(Settings::getUserValue("Language", "en").toString());
			qApp->processEvents();
			QTimer::singleShot(1000, manager, SLOT(init()));
			ret = qApplication.exec();
			delete manager;
#ifdef Q_OS_LINUX
		}
#endif
	}

	return ret;
}
