#include "loginscreen.h"

LoginScreen::LoginScreen(QWidget *parent ) : QWidget( parent )
{
	screen = QGuiApplication::screens()[0];
	setGeometry(screen->geometry());
	QIcon icon(":/icons/other/librework.icns");
	setWindowIcon( icon );
	setWindowTitle(tr("Librework Desktop"));
	setWindowFlags( windowFlags() | Qt::FramelessWindowHint );

	manager = 0;
#ifdef Q_OS_LINUX
	clientSession = 0;
#endif

	qsrand( QTime::currentTime().msec() );

	initInitScreen();
	initUserControls();
	showUserControls( false );

	connect(passwordLineEdit, SIGNAL(returnPressed()), this, SLOT(loginButtonClicked()));
	connect(usernameLineEdit, SIGNAL(returnPressed()), this, SLOT(loginButtonClicked()));
	connect(loginButton, SIGNAL(clicked(bool)), this, SLOT(loginButtonClicked()));
	connect(exitButton, SIGNAL(customContextMenuRequested(QPoint)), exitButton, SLOT(showMenu()));

	pixmapFrame->show();
	pixmapFrame->setFocusPolicy( Qt::StrongFocus );
	pixmapFrame->setFocus();

	connect( animation, SIGNAL( finished() ), this, SLOT( animationFinished() ) );
	connect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( doAnimation() ) );
	connect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( doAnimation() ) );
	connect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( doAnimation(QKeyEvent*) ) );
	connect( pixmapFrame, SIGNAL( escapePressed() ), this, SLOT( doInitScreen() ) );

	QTimer::singleShot( 250, this, SLOT( setDateTime() ) );

	timer = new QTimer();
	timer->setInterval( 1000 );
	connect( timer, SIGNAL( timeout() ), this, SLOT( setDateTime() ) );
	timer->start();
}

LoginScreen::~LoginScreen()
{
	timer->stop();
	animation->stop();
	delete animation;
	delete timer;
	delete pixmapFrame;
	delete logoLabel;
	delete usernameLineEdit;
	delete passwordLineEdit;
	delete serverComboBox;
	delete loginButton;
	delete exitButton;
	if ( manager != 0 )
	{
		delete manager;
	}
#ifdef Q_OS_LINUX
	if ( clientSession != 0 )
	{
		delete clientSession;
	}
#endif
}

void LoginScreen::initInitScreen()
{
	QString configurationFile = selectConfigurationFile();

	QPixmap pixmap = selectPixmap( configurationFile );

	pixmapFrame = new PixmapFrame( this, pixmap );
	pixmapFrame->setColor( QColor(0,0,0) );
	int pX = 5;
	int pY = 75;

	animation = new QPropertyAnimation(pixmapFrame, "opacity");
	animation->setDuration(600);
	animation->setStartValue( 1.0 );
	animation->setEndValue( LOGINSCREEN_OBSCURED_OPACITY );

	if ( ! configurationFile.isEmpty() )
	{
		QSettings settings( configurationFile, QSettings::IniFormat );
		pixmapFrame->setFontColor( QColor( settings.value( "LoginScreen/FontColor" , "#ffffff" ).toString() ) );
		pX = settings.value( "LoginScreen/TextPositionX", pX ).toInt();
		pY = settings.value( "LoginScreen/TextPositionY", pY ).toInt();
	}
	pixmapFrame->setTextPos( pX * screen->geometry().width() / 100, pY * screen->geometry().height() / 100 );

	pixmapFrame->setFixedSize( screen->geometry().size() );

	pixmapFrame->setFontFamily( LIBREWORK_FONT );
	pixmapFrame->setTimeFontPixelSize( screen->geometry().width() / 15);
	pixmapFrame->setDateFontPixelSize( screen->geometry().width() / 25);
	pixmapFrame->resize();
}

void LoginScreen::initUserControls()
{
	QSize desktopSize = screen->geometry().size();

	QSize logoLabelSize = QSize( desktopSize.width() / 10, desktopSize.width() / 10 );
	QPoint logoLabelPos = QPoint( ( desktopSize.width() - logoLabelSize.width() ) / 2, desktopSize.height() * 20 / 100 );
	QString logoLabelStyleSheet(
		"QLabel"
		"{"
		"	background-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 0px;"
		"	border-style: solid;"
		"	border-radius: 20px;"
		"}"
		);	
	logoLabel = new QLabel( pixmapFrame );
	logoLabel->setFocusPolicy( Qt::NoFocus );
	logoLabel->setStyleSheet( logoLabelStyleSheet );
	logoLabel->setScaledContents( true );
	logoLabel->setFixedSize( logoLabelSize );
	logoLabel->setPixmap( QPixmap( ":/icons/other/librework.png" ) );
	logoLabel->setGeometry( logoLabelPos.x(), logoLabelPos.y(), logoLabelSize.width(), logoLabelSize.height() );

	QSize usernameLineEditSize( desktopSize.width() / 4, desktopSize.height() / 15 );

	int spacing = desktopSize.height() / 50;
	QPoint usernameLineEditPos( ( desktopSize.width() - usernameLineEditSize.width() ) / 2, logoLabelPos.y() + logoLabelSize.height() + spacing );
	usernameLineEdit = new QLineEdit( pixmapFrame );
	usernameLineEdit->setPlaceholderText( tr("User") );
	usernameLineEdit->setAlignment(Qt::AlignCenter);
	usernameLineEdit->setFixedSize( usernameLineEditSize );
	usernameLineEdit->setGeometry( usernameLineEditPos.x(), usernameLineEditPos.y(), usernameLineEditSize.width(), usernameLineEditSize.height() );
	QString lastLogin = Settings::getValue("LastLogin"); // 'General' section must not be specified
	usernameLineEdit->setText( lastLogin );
	usernameLineEditTextChanged( lastLogin );
	connect( usernameLineEdit, SIGNAL(textChanged(QString)), this, SLOT(usernameLineEditTextChanged(QString)));

	QSize passwordLineEditSize( desktopSize.width() / 4, desktopSize.height() / 15 );
	QPoint passwordLineEditPos( ( desktopSize.width() - passwordLineEditSize.width() ) / 2, usernameLineEditPos.y() + usernameLineEditSize.height() + spacing );
	passwordLineEdit = new QLineEdit( pixmapFrame );
	passwordLineEdit->setPlaceholderText( tr("Password") );
	passwordLineEdit->setAlignment(Qt::AlignCenter);
	QString passwordStyleSheet = QString(
		"QLineEdit"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 3px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	font: 75 Bold %1px '%2';"
		"	color: #ededed;"
		"}"
		).arg(desktopSize.width() / 64).arg( LIBREWORK_FONT );
	passwordLineEdit->setEchoMode(QLineEdit::Password);
	passwordLineEdit->setStyleSheet( passwordStyleSheet );
	passwordLineEdit->setFixedSize( passwordLineEditSize );
	passwordLineEdit->setGeometry( passwordLineEditPos.x(), passwordLineEditPos.y(), passwordLineEditSize.width(), passwordLineEditSize.height() );

	QSize serverComboBoxSize( desktopSize.width() / 6, desktopSize.height() / 25 );
	QPoint serverComboBoxPos( ( desktopSize.width() - serverComboBoxSize.width() ) / 2, passwordLineEditPos.y() + passwordLineEditSize.height() + spacing );
	serverComboBox = new QComboBox( pixmapFrame );
	if ( Settings::getValue( "Server/LibreworkServerPort").toInt() > 0 )
	{
		serverComboBox->addItem( QIcon( ":/icons/32x32/start.png" ), Settings::getValue( "Server/LibreworkServerAddress"), false );
	}
	if ( Settings::getValue( "Desktop/AllowLocalDesktopLogin").toLower() == "true" )
	{
		serverComboBox->addItem( QIcon( ":/icons/32x32/start.png" ), tr("This Computer"), false );
	}
#ifdef Q_OS_LINUX
	if ( Settings::getValue( "Desktop/AllowRemoteDesktopLogin").toLower() == "true" )
	{
		ClientSettings * clientSettings = new ClientSettings();
		QStringList servers = clientSettings->serverList() + clientSettings->userServerList();
		delete clientSettings;
		for (int i = 0; i < servers.count(); i++ )
		{
			serverComboBox->addItem( QIcon( ":/icons/32x32/desktop.png" ), servers[i], true );
		}
	}
#endif
	if ( serverComboBox->count() == 0 )
	{
		serverComboBox->addItem( QIcon( ":/icons/32x32/start.png" ), tr("This Computer"), false );
	}

	QString lastServer = Settings::getValue("LastServer"); // 'General' section must not be specified
	if ( Settings::getBooleanValue( "RememberServers") ) // 'General' section must not be specified, again
	{
		QStringList macs = Settings::allKeys( CONFIG_FILE, "Default Servers");
		QString macAddress = getMacAddress();
		if ( macs.contains( macAddress, Qt::CaseInsensitive) )
		{
			lastServer = Settings::getValue("Default Servers/" + macAddress);
		}
	}

	connect( serverComboBox, SIGNAL(currentTextChanged(QString)), this, SLOT(serverChanged(QString)));

	qDebug("LastServer: '%s'", lastServer.toUtf8().data());
	if ( serverComboBox->findText( lastServer, Qt::MatchExactly ) != -1 )
	{
		serverComboBox->setCurrentText( lastServer );
	}
	QFont arial( LIBREWORK_FONT );
	arial.setPixelSize( desktopSize.width() / 60 );
	serverComboBox->setFont( arial );
	QString serverStyleSheet = QString(
		"QComboBox"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0 );"
		"	border-width: 0px;"
		"	border-style: solid;"
		"	border: none;"
		"	selection-background-color: rgba( 127, 127, 127, 0.5);"
		"	font: 75 Bold %1px '%2';"
		"	font-size: %1px;"
		"	icon-size: %1px;"
		"	border-radius: 8px;"
		"	color: #ededed;"
		"}"
		).arg(desktopSize.width() / 72).arg( LIBREWORK_FONT );
	serverComboBox->setStyleSheet( serverStyleSheet );
	serverComboBox->setFixedSize( serverComboBoxSize );
	serverComboBox->setGeometry( serverComboBoxPos.x(), serverComboBoxPos.y(), serverComboBoxSize.width(), serverComboBoxSize.height() );

	QString loginScreenButtonStyle = QString(
		"QPushButton"
		"{"
		"	background-color: transparent;"
		"	border: none;"
		"	icon-size: %1px;"
		"}"
		"QPushButton:selected"
		"{"
		"	background-color: transparent;"
		"	border: none;"
		"	icon-size: %1px;"
		"}"
		"QPushButton:pressed:hover"
		"{"
		"	background-color: transparent;"
		"	border: none;"
		"       padding-top: 5px;"
		"       padding-left: 5px;"
		"	icon-size: %1px;"
		"}"
		"QPushButton:hover"
		"{"
		"	background-color: rgba( 255, 255, 255, 25 );"
		"	border-width: 0px;"
		"	border-style: solid;"
		"	border-radius: %2px;"
		"	icon-size: %1px;"
		"};"
		).arg(desktopSize.width() / 40).arg(desktopSize.height() / 10);
	QSize buttonSize( desktopSize.width() / 25, desktopSize.width() / 25 );

	QPoint loginButtonPos( passwordLineEditPos.x() + passwordLineEditSize.width() + 5, passwordLineEditPos.y() );
	loginButton = new QPushButton( QIcon( ":/icons/64x64/logon.png" ), "", pixmapFrame );
	loginButton->setFocusPolicy( Qt::NoFocus );
	loginButton->setFlat( true );
	loginButton->setStyleSheet( loginScreenButtonStyle );
	loginButton->setGeometry( loginButtonPos.x(), loginButtonPos.y(), buttonSize.width(), buttonSize.height() );

	QPoint exitButtonPos( desktopSize.width() - buttonSize.width() - spacing, desktopSize.height() - buttonSize.height() - spacing );
	exitButton = new QPushButton( QIcon( ":/icons/64x64/logoff.png" ), "", pixmapFrame );
	exitButton->setFocusPolicy( Qt::NoFocus );
	exitButton->setFlat( true );
	exitButton->setStyleSheet( loginScreenButtonStyle );
	exitButton->setGeometry( exitButtonPos.x(), exitButtonPos.y(), buttonSize.width(), buttonSize.height() );
	setShutdownMenu();
}

void LoginScreen::serverChanged(QString server)
{
	ClientSettings * clientSettings = new ClientSettings();
	usernameLineEdit->setText( clientSettings->getLogin( server ) );
	delete clientSettings;
}

void LoginScreen::usernameLineEditTextChanged(QString text)
{
	QSize desktopSize = screen->geometry().size();
	int borderWidth = 0;
	if ( text.isEmpty() )
	{
		borderWidth = 3;
	}

	QString usernameStyleSheet = QString(
		"QLineEdit"
		"{"
		"	background-color: transparent;"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: %1px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	font: 75 Bold %2px '%3';"
		"	color: #ededed;"
		"}"
		"QLineEdit:focus"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 3px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	font: 75 Bold %2px '%3';"
		"	color: #ededed;"
		"}"
		).arg(borderWidth).arg(desktopSize.width() / 48).arg( LIBREWORK_FONT );
	usernameLineEdit->setStyleSheet( usernameStyleSheet );
}

void LoginScreen::showUserControls(bool show)
{
	if ( show )
	{
		logoLabel->show();
		usernameLineEdit->show();
		passwordLineEdit->show();
		serverComboBox->show();
		if ( ! passwordLineEdit->text().isEmpty() )
		{
			loginButton->show();
		}
		exitButton->show();
		connect( passwordLineEdit, SIGNAL(textChanged(const QString &)), this, SLOT(passwordTextChanged(const QString &) ));
	} else {
		logoLabel->hide();
		usernameLineEdit->hide();
		passwordLineEdit->hide();
		serverComboBox->hide();
		loginButton->hide();
		exitButton->hide();
		disconnect( passwordLineEdit, SIGNAL(textChanged(const QString &)), this, SLOT(passwordTextChanged(const QString &) ));
	}
}

void LoginScreen::passwordTextChanged(const QString & )
{
	if ( ! passwordLineEdit->text().isEmpty() )
	{
		loginButton->show();
	} else {
		loginButton->hide();
	}
}

void LoginScreen::doAnimation(QKeyEvent * event)
{
	pixmapFrame->setFocusPolicy( Qt::NoFocus );
	passwordLineEdit->clear();
	disconnect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( doAnimation() ) );
	disconnect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( doAnimation() ) );
	disconnect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( doAnimation(QKeyEvent*) ) );
	timer->stop();

	showUserControls( true );

	if ( event != 0 )
	{
		LoginScreen::keyPressEvent( event );
	}

	if ( usernameLineEdit->text().isEmpty() )
	{
		usernameLineEdit->setFocus();
	} else {
		passwordLineEdit->setFocus();
	}

	animation->start();
}

void LoginScreen::doInitScreen()
{
	pixmapFrame->setFocusPolicy( Qt::StrongFocus );
	pixmapFrame->setFocus();
	showUserControls( false );
	animation->stop();
	pixmapFrame->setOpacity( 1.0 );
	pixmapFrame->showText( true );
	timer->start();
	connect( pixmapFrame, SIGNAL( mousePressed() ), this, SLOT( doAnimation() ) );
	connect( pixmapFrame, SIGNAL( mouseMoved() ), this, SLOT( doAnimation() ) );
	connect( pixmapFrame, SIGNAL( keyPressed(QKeyEvent*) ), this, SLOT( doAnimation(QKeyEvent*) ) );
}

void LoginScreen::animationFinished()
{
	pixmapFrame->showText( false );
}

QString LoginScreen::selectConfigurationFile()
{
	QString ret;
	QDir wallpaperDir( SHAREDIR + QString(QDir::separator()) + "/loginscreen" );
	QStringList wallpaperConfigList = wallpaperDir.entryList( QStringList( "*.ini" ) );
	if ( wallpaperConfigList.count() > 0 )
	{
		int index = qrand() % wallpaperConfigList.count();
		ret = wallpaperDir.absolutePath() + QDir::separator() + wallpaperConfigList[index];
	}
	return ret;
}

void LoginScreen::setDateTime()
{
	QString time = QTime::currentTime().toString( "hh:mm" );
	QString date = QDate::currentDate().toString( "dddd, d MMM" );
	date.replace( 0, 1, date[0].toUpper() );
	pixmapFrame->setText( time, date );
}

QPixmap LoginScreen::selectPixmap( QString configurationFile)
{
	QPixmap pixmap;
	if ( ! configurationFile.isEmpty() )
	{
		QSettings settings( configurationFile, QSettings::IniFormat );
		QRect screenSize = screen->geometry();
		QString sizeString = QString::number( screenSize.width() ) + "x" + QString::number( screenSize.height() );
		if ( settings.value( sizeString + "/File", "").toString().isEmpty() ) // In case not exact resolution found
		{
			// Find best aspect ratio file
			double aspectRatio = (double) screenSize.width() / (double) screenSize.height();
			double bestAspectRatioDiff = -1;
			QStringList childGroups = settings.childGroups();
			for ( int i = 0; i < childGroups.count(); i++ )
			{
				bool ok;
				if ( childGroups[i] != "LoginScreen" )
				{
					int width = childGroups[i].section("x", 0, 0).toInt(&ok);
					if ( ! ok )
					{
						break;
					}
					int height = childGroups[i].section("x", 1, 1).toInt(&ok);
					if ( ok && width >= screenSize.width() && height >= screenSize.height() )
					{
						double pixmapAspectRatio = (double) width / (double) height;
						double diff = pixmapAspectRatio - aspectRatio;
						if ( diff < 0 )
						{
							diff = -diff;
						}
						if ( diff < bestAspectRatioDiff || bestAspectRatioDiff == -1 )
						{
							bestAspectRatioDiff = diff;
							sizeString = childGroups[i];
						}
					}
				}
			}
		}
		QString pixmapFile = settings.value( sizeString + "/File", settings.value( "LoginScreen/File", "" ) ).toString();
		if ( ! pixmapFile.isEmpty() )
		{
			pixmap = QPixmap( SHAREDIR + QString( QDir::separator() ) + "loginscreen" + QString( QDir::separator() ) + pixmapFile );
		}
	} else {
		QDir wallpaperDir( SHAREDIR + QString(QDir::separator()) + "/loginscreen" );
		QStringList extensions;
		extensions << "*.png";
		extensions << "*.jpg";
		extensions << "*.gif";
		QStringList wallpaperList = wallpaperDir.entryList( extensions );
		if ( wallpaperList.count() > 0 )
		{
			int index = qrand() % wallpaperList.count();
			pixmap = QPixmap( wallpaperDir.absolutePath() + QDir::separator() + wallpaperList[index] );
		}
	}
	return pixmap;
}

void LoginScreen::setShutdownMenu()
{
	QSize desktopSize = screen->geometry().size();
	QMenu * menu = new QMenu;
	menu->setStyle( new MenuStyle(desktopSize.width() / 100) );
	menu->addAction(QIcon(":/icons/32x32/logoff.png"), tr("Power Off"), Power::instance(), SLOT(powerOff()));
	menu->addAction(QIcon(":/icons/32x32/reboot.png"), tr("Reboot"), Power::instance(), SLOT(reboot()));
	menu->addAction(QIcon(":/icons/32x32/suspend.png"), tr("Suspend"), Power::instance(), SLOT(suspend()));
	menu->addAction(QIcon(":/icons/32x32/hibernate.png"), tr("Hibernate"), Power::instance(), SLOT(hibernate()));
	int paddingLeft = desktopSize.width() / 60;
	int paddingRight = desktopSize.width() / 75;
	int menuHeight = desktopSize.height() / 30;
	int fontPixelSize = desktopSize.height() / 50;
	QString menuStyleSheet = QString(
		"QMenu"
		"{"
		"	background-color: rgba( 0, 0, 0, 0.5 );"
		"	border-color: rgba( 255, 255, 255, 0.5 );"
		"	border-width: 3px;"
		"	border-style: solid;"
		"	border-radius: 8px;"
		"	padding: 10px;"
		"	font: 75 Bold %2px '%3';"
		"	color: #ededed;"
		"}"
		"QMenu::item"
		"{"
		"	height: %1px;"
		"	margin: 0px;"
		"	padding-left: %4px;"
		"	padding-right: %5px;"
		"}"
		"QMenu::item:selected"
		"{"
		"	background-color: rgba( 127, 127, 127, 0.5 );"
		"}"
		"QMenu::item:hover"
		"{"
		"	background-color: rgba( 127, 127, 127, 0.5 );"
		"}"
		).arg( menuHeight ).arg( fontPixelSize ).arg( LIBREWORK_FONT ).arg(paddingLeft).arg(paddingRight);
	menu->setStyleSheet( menuStyleSheet );
	exitButton->setMenu( menu );
}

void LoginScreen::loginButtonClicked()
{
	if ( usernameLineEdit->text().isEmpty() || passwordLineEdit->text().isEmpty() )
	{
		return;
	}

	setCursor( Qt::WaitCursor );
	pixmapFrame->setOpacity( LOGINSCREEN_OBSCURED_OPACITY );
	qApp->processEvents();

	InitProgress * initProgress = new InitProgress(); // class Desktop will take care of its deletion
	initProgress->setCursor( Qt::WaitCursor );
	showUserControls( false );
	initProgress->showMinimized();
	qApp->processEvents();
	QScreen * initProgressScreen = initProgress->windowHandle()->screen();
	initProgress->setGeometry( ( initProgressScreen->geometry().width() - INITPROGRESS_WIDTH ) / 2, ( initProgressScreen->geometry().height() - INITPROGRESS_HEIGHT ) / 2, INITPROGRESS_WIDTH, INITPROGRESS_HEIGHT );
	initProgress->showNormal();
	qApp->processEvents();

	QPropertyAnimation * fadeOut = new QPropertyAnimation( this ); // LoginScreen is its parent, so it takes care of its deletion
	fadeOut->setTargetObject( pixmapFrame );
	fadeOut->setPropertyName( "opacity" );
	fadeOut->setDuration(1000);
	fadeOut->setStartValue( LOGINSCREEN_OBSCURED_OPACITY );
	fadeOut->setEndValue( 0.0 );
	connect( fadeOut, SIGNAL(finished()), this, SLOT(hide()));
	fadeOut->start();
	qApp->processEvents();

	qDebug("Setting last server: '%s'", serverComboBox->currentText().toUtf8().data());
	Settings::setValue(CONFIG_FILE, "LastServer", serverComboBox->currentText()); // 'General' section must not be specified
	if ( Settings::getBooleanValue("RememberServers") )
	{
		QString macAddress = getMacAddress();
		Settings::setValue(CONFIG_FILE, "Default Servers/" + macAddress, serverComboBox->currentText() );
	}

	if ( serverComboBox->currentData().toBool() ) // is remote session
	{
#ifdef Q_OS_LINUX
		initProgress->setProgress( 70 );
		qApp->processEvents();
		clientSession = new ClientSession();
		connect( clientSession, SIGNAL(finished()), this, SLOT(show()));
		connect( clientSession, SIGNAL(finished()), this, SLOT(doInitScreen()));
		clientSession->launchProcess(usernameLineEdit->text(), passwordLineEdit->text(), serverComboBox->currentText() );
		initProgress->close();
		delete initProgress;
#endif
	} else { // is local session or local session against Librework server
		QStringList error = Session::launch( usernameLineEdit->text(), passwordLineEdit->text(), serverComboBox->currentText() == tr("This Computer"), this, manager, initProgress );
		if ( error.count() == 2 )
		{
			setCursor( Qt::ArrowCursor );
			initProgress->close();
			delete initProgress;
			qApp->processEvents();
			doAnimation();
			show();
			QMessageBox::critical( 0, error[0], error[1] );
		} else {
			passwordLineEdit->clear();
			connect( fadeOut, SIGNAL(finished()), manager, SLOT(init()));
		}
	}
	setCursor( Qt::ArrowCursor );
	qApp->processEvents();
}

QString LoginScreen::getMacAddress()
{
	foreach(QNetworkInterface netInterface, QNetworkInterface::allInterfaces())
	{
		if ( ! (netInterface.flags() & QNetworkInterface::IsLoopBack) )
		{
			return netInterface.hardwareAddress();
		}
	}
	return QString();
}

void LoginScreen::forceTranslations()
{
	QString ok = tr("OK");
	Q_UNUSED(ok);
	QString cancel = tr("Cancel");
	Q_UNUSED(cancel);
}
