#include "manager.h"

#include <QCoreApplication>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

Manager::Manager(Desktop * parent)
{
	desktop = parent;
	
	qApp->installNativeEventFilter( this );
	qDebug() << "Installed native event filter";

	XSelectInput(QX11Info::display(), QX11Info::appRootWindow(QX11Info::appScreen()),
		SubstructureRedirectMask /* MapRequest, ConfigureRequest, CirculateRequest */
		| SubstructureNotifyMask /* CreateNotify, DestroyNotify, MapNotify, UnmapNotify,
						ReparentNotify, GravityNotify, ConfigureNotify, CirculateNotify */
		| ButtonPressMask );     /* ButtonPress */
	XFlush(QX11Info::display());

	instance.append(this);
}

Manager::~Manager()
{
	instance.removeOne(this);
}

bool Manager::nativeEventFilter(const QByteArray & eventType, void * message, long * result)
{
	Q_UNUSED(result);

	bool res = false;

	if ( eventType == "xcb_generic_event_t" )
	{
		xcb_generic_event_t *xcbEvent = (xcb_generic_event_t *)message;

		switch (xcbEvent->response_type & ~0x80)
		{
			case XCB_EXPOSE:
				//qDebug("Expose event on window %d", ((xcb_expose_event_t *)xcbEvent)->window);
				break;
			case XCB_CREATE_NOTIFY:
				//qDebug() << "Create notify event";
				res = createNotifyEvent( xcbEvent );
				break;
			case XCB_MAP_NOTIFY:
				//qDebug("Map notify event on window %d", ((xcb_map_notify_event_t *)xcbEvent)->window);
				res = mapNotifyEvent( xcbEvent );
				break;
			case XCB_MAP_REQUEST:
				//qDebug() << "Map request event";
				res = mapRequestEvent( xcbEvent );
				break;
			case XCB_DESTROY_NOTIFY:
				//qDebug() << "Destroy notify event";
				res = destroyNotifyEvent( xcbEvent );
				break;
			case XCB_UNMAP_NOTIFY:
				//qDebug() << "Unmap notify event";
				res = unmapNotifyEvent( xcbEvent );
				break;
			case XCB_CONFIGURE_REQUEST:
				//qDebug() << "Configure request event";
				res = configureRequestEvent( xcbEvent );
				break;
			case XCB_PROPERTY_NOTIFY:
				//qDebug() << "Property notify event";
				res = propertyNotifyEvent( xcbEvent );
				break;
			case XCB_BUTTON_PRESS:
				//qDebug() << "Button press event";
				res = buttonPressEvent( xcbEvent );
				break;
			default:
				//qDebug("Unknown event type: %d", xcbEvent->response_type);
				break;
		}
	}

	return res;
}

bool Manager::propertyNotifyEvent( xcb_generic_event_t * event )
{
	xcb_property_notify_event_t * propertyNotifyEvent = (xcb_property_notify_event_t *) event;
	if( propertyNotifyEvent->atom == XCB_ATOM_WM_NAME )
	{
		emit titleChanged( (long long)propertyNotifyEvent->window, X11Info::getWindowName( propertyNotifyEvent->window ) );
	}

	return false;
}

bool Manager::buttonPressEvent( xcb_generic_event_t * event )
{
	xcb_button_press_event_t * buttonEvent = (xcb_button_press_event_t *) event;
	Window win = buttonEvent->event;
	qDebug("Button press event on window %d, parent %d", (int) win, (int) X11Info::getParent(win));

	X11Info::setInputFocus( X11Info::getParent(win) );
	X11Info::setInputFocus( win );
	
	return false;
}

bool Manager::mapRequestEvent( xcb_generic_event_t * event )
{
	xcb_map_request_event_t * mapRequestEvent = (xcb_map_request_event_t *) event;
	Window win = mapRequestEvent->window;

	XSelectInput(QX11Info::display(), win, PropertyChangeMask);

	qDebug("Trying to get window id %d", (int)win);
	Settings::logInfo(QString("Window name: %1, PID: %2, type: %3").arg( X11Info::getWindowName(win)).arg( X11Info::getWindowPid(win)).arg( X11Info::getWindowType(win)));

	QString windowType = X11Info::getWindowType(win);

	if (windowType.startsWith("Unknown"))
	{
		QSize max = X11Info::getWindowMaxSize(win);

		if (max == X11Info::getWindowMinSize(win) && max != QSize(0,0))
		{
			windowType = "Dialog";
		}
	}

	if (windowType == "Splash")
	{
		qDebug("Window type splash");
		mapSplash(win);
	}
	else if (windowType == "Dialog")
	{
		qDebug("Window type dialog");
		mapDialog(win);
	}
	else if (windowType == "Normal")
	{
		qDebug("Window type normal");
		showWindow(win);
	}
	else // Default action
	{
		qDebug("Window type default");
		showWindow(win);
	}

	return true;
}

bool Manager::createNotifyEvent( xcb_generic_event_t * event)
{
	xcb_create_notify_event_t * createNotifyEvent = (xcb_create_notify_event_t *) event;
	Window window = createNotifyEvent->window;
	Settings::logInfo( QString("Create notify: window: %1, save_under: %2, type: %3").arg(window).arg( X11Info::getWindowSaveUnder(window)).arg( X11Info::getWindowType(window)));

	bool isDialog = ( X11Info::getWindowType(window) == "Dialog" );
	if ( isDialog && X11Info::getWindowMapState(window) == IsViewable )
		resizeWindow( window, isDialog );

	return true;
}

bool Manager::mapNotifyEvent( xcb_generic_event_t * event)
{
	xcb_map_notify_event_t * mapNotifyEvent = (xcb_map_notify_event_t *) event;
	Window window = mapNotifyEvent->window;
	Settings::logInfo( QString("Map notify: window: %1").arg(window));

	if ( mapNotifyEvent->override_redirect != 1 )
	{
		emit mapNotify( (long long) window );
	}

	return false;
}
bool Manager::configureRequestEvent( xcb_generic_event_t * event )
{
	xcb_configure_request_event_t * configureRequestEvent = (xcb_configure_request_event_t *) event;
	Window window = configureRequestEvent->window;
	QSize preferredSize( configureRequestEvent->width, configureRequestEvent->height );
	bool isDialog = ( X11Info::getWindowType(window) == "Dialog" );

	int width = configureRequestEvent->value_mask & CWWidth;
	int height = configureRequestEvent->value_mask & CWHeight;
	bool hasWidth = ( width == (int) CWWidth );
	bool hasHeight = ( height == (int) CWHeight );
	//qDebug("ConfigureRequest event on window %d, value_mask: %d, size: %dx%d, width: %d == %d, height: %d == %d", (int) window, configureRequestEvent->value_mask, configureRequestEvent->width, configureRequestEvent->height, width, (int)CWWidth, height, (int)CWHeight);
	if ( hasWidth && hasHeight )
	{
		resizeWindow( window, isDialog, preferredSize );
	}

	return true;
}

bool Manager::unmapNotifyEvent( xcb_generic_event_t * event )
{
	xcb_unmap_notify_event_t * unmapNotifyEvent = (xcb_unmap_notify_event_t *) event;

	//qDebug("Unmapnotify: Trying to close window %d with signal %d", (int)unmapNotifyEvent->window, unmapNotifyEvent->response_type );
	//Settings::logInfo( QString("Unmapnotify: Trying to close window %1 with signal %2")
        //               .arg(unmapNotifyEvent->window)
        //               .arg(unmapNotifyEvent->response_type));

	Decorator::closeDecorator( unmapNotifyEvent->window );

    emit unmapWindow( (long long)unmapNotifyEvent->window );

	return true;
}

bool Manager::destroyNotifyEvent( xcb_generic_event_t * event )
{
	xcb_destroy_notify_event_t * destroyNotifyEvent = (xcb_destroy_notify_event_t *) event;
	//qDebug("DestroyNotify: Trying to close window %d with signal %d", (int)destroyNotifyEvent->window, destroyNotifyEvent->response_type);
	Settings::logInfo( QString("DestroyNotify: Trying to close window %1 with signal %2")
                        .arg((int)destroyNotifyEvent->window)
                        .arg(destroyNotifyEvent->response_type));
	Decorator::closeDecorator( destroyNotifyEvent->window );

    emit unmapWindow( (long long)destroyNotifyEvent->window );

	return true;
}
void Manager::resizeWindow( Window window, bool isDialog, QSize preferredSize )
{
	qDebug("Resizing window %d, preferredSize: %dx%d, base size: %dx%d, max size:%dx%d, min size: %dx%d, isDialog: %d", (int) window,
			preferredSize.width(), preferredSize.height(),
			X11Info::getWindowBaseSize(window).width(), X11Info::getWindowBaseSize(window).height(),
			X11Info::getWindowMaxSize(window).width(), X11Info::getWindowMaxSize(window).height(),
			X11Info::getWindowMinSize(window).width(), X11Info::getWindowMinSize(window).height(),
			isDialog
	);
	int width, height;
	int posx, posy;

	QSize max = X11Info::getWindowMaxSize(window);
	QSize min = X11Info::getWindowMinSize(window);
	QSize base = X11Info::getWindowBaseSize(window);
	QSize pref;

	Settings::logInfo( QString("Resize window. Max: %1x%2, Min: %3x%4, Base: %5x%6")
			.arg(max.width())
			.arg(max.height())
			.arg(min.width())
			.arg(min.height())
			.arg(base.width())
			.arg(base.height()) );

	if ( max == min  && min == base  && base == QSize(0, 0) )
		return;

	if ( preferredSize == QSize(0, 0) )
	{
		pref = X11Info::getWindowBaseSize(window);
	} else {
		pref = preferredSize;
	}

	QSize screenSize = desktop->getDesktopSize();

	//if ( ( max == min && max != QSize(0,0) ) || isDialog )
	if ( isDialog )
	{
		width = pref.width();
		height = pref.height();
		posx = (screenSize.width() - width) / 2;
		posy = (screenSize.height() - height) /2;
		decorateWindow( window, pref, QPoint(posx, posy) );
		XMoveResizeWindow(QX11Info::display(), window, posx, posy, width, height);
		qDebug("Resized to %dx%d.", width, height);
		X11Info::setInputFocus( window );
	} else {
		if (max != QSize(0,0) || min != QSize(0,0) || base != QSize(0,0))
		{
			if ( pref.width() <= screenSize.width() )
			{
				width = pref.width();
				posx = (screenSize.width() - width) / 2;
			} else {
				width = screenSize.width();
				posx = 0;
			}
			if ( pref.height() <= screenSize.height() )
			{
				height = pref.height();
				posy = (screenSize.height() - height) /2;
			} else {
				height = screenSize.height();
				posy = 0;
			}

			decorateWindow( window, pref, QPoint(posx, posy) );
			XMoveResizeWindow(QX11Info::display(), window, posx, posy, width, height);
			qDebug("Resized to %dx%d.", width, height);
			X11Info::setInputFocus( window );
		}
	}
}

void Manager::decorateWindow(Window window, QSize wsize, QPoint pos)
{
	if ( X11Info::getWindowMapState(window) == IsViewable )
	{
		qDebug("Decorating window");
		if ( Decorator::hasDecorator( window ) )
			Decorator::closeDecorator( window );
		new Decorator(pos, wsize, window, 0);
	}
}

void Manager::mapSplash(Window window)
{
	XMapWindow(QX11Info::display() , window );
	XSync(QX11Info::display(), false);
}

void Manager::mapDialog(Window window)
{
	qDebug("Trying to show dialog window %d", (int)window );
	XMapRaised(QX11Info::display() , window );

	resizeWindow( window, true, X11Info::getWindowPreferredSize( window ) );

	X11Info::setInputFocus( window );
	XSync(QX11Info::display(), false);
}

void Manager::showWindow(Window window)
{
	QString appName = X11Info::getWindowName(window);
	//QIcon icon = X11Info::getWindowIcon(window);
	QIcon icon = currentIcon;
	currentIcon = QIcon(ICON_DIR_APPS + QString(QDir::separator()) + "default.png");
	QString name = currentName;
	//Q_PID pid = X11Info::getWindowPid(window);
	X11Info::setInputFocus( window );

	qDebug("Trying to attach window %d", (int)window);
	X11Info::attachWindow(window);
	emit mapWindow( (long long)window );
}

void Manager::attachWindow(QWidget * container, Window window)
{
	X11Info::attachWindow( container, window );
	emit fileOpened();
}
