#include "session.h"
#include "manager.h"

#ifdef Q_OS_LINUX
static pam_handle_t *local_auth_handle = NULL; // this gets set by pam_start
struct pam_response * reply;
#endif

void Session::getUserAndGroupIds( QString username, int * uid, int * gid )
{
#ifdef Q_OS_WIN
	Q_UNUSED( username );
	Q_UNUSED( uid );
	Q_UNUSED( gid );
#endif
#ifdef Q_OS_LINUX
	struct passwd pwd;
	struct passwd *result;
	char *buf;
	size_t bufsize;
	int s;

	bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (bufsize == (size_t)-1)          /* Value was indeterminate */
	{
		bufsize = 16384;        /* Should be more than enough */
	}

	buf = (char*)malloc(bufsize);
	if (buf == NULL)
	{
		* uid = -1;
		* gid = -1;
		qDebug("Error getting memory for buffer");
		return;
	}

	s = getpwnam_r(username.toUtf8().data(), &pwd, buf, bufsize, &result);
	if (result == NULL)
	{
		if (s == 0)
		{
			qDebug("Not found\n");
		} else {
			qDebug("Error %d on getpwnam_r", s);
		}
		* uid = -1;
		* gid = -1;
	} else {
		* uid = result->pw_uid;
		* gid = result->pw_gid;
	}
	free( buf );
#endif
} 

int Session::setCurrentDir(const char *dirname)
{
	QDir::setCurrent(dirname);
#ifdef Q_OS_WIN
	if (SetCurrentDirectoryA(dirname))
	{
		return 0;
	} else {
		return -1;
	}
#endif

#ifdef Q_OS_LINUX
	return chdir(dirname);
#endif
}

bool Session::createSystemUser( int uid, QString homes, QString username )
{
#ifdef Q_OS_WIN
	Q_UNUSED( uid );
	Q_UNUSED( homes );
	Q_UNUSED( username );
#endif
#ifdef Q_OS_LINUX
	bool ret = false;
	bool exists = false;
	QFile passwd("/etc/passwd");
	if (passwd.open(QIODevice::ReadOnly))
	{
		QTextStream in(&passwd);
		while (!in.atEnd())
		{
			QString line = in.readLine();
			QString lineUsername = line.section(":", 0, 0);
			int lineUserId = line.section(":", 2, 2).toInt();
			if ( lineUserId == uid )
			{
				exists = true;
				if ( lineUsername == username )
				{
					ret = true;
				}
			}
		}
		passwd.close();
	}

	if ( ! exists )
	{
		QFile passwd("/etc/passwd");
		passwd.open(QIODevice::Append | QIODevice::Text);
		QByteArray line;
		line += username;
		line += ":!:"; // No password entry
		line += QString::number( uid );
		line += ":50::"; // Staff group
		line += homes + "/" + username;
		line += ":/bin/bash";
		if ( passwd.write( line ) == -1 )
		{
			ret = false;
		} else {
			ret = true;
		}
		passwd.close();
	}
	return ret;
#endif
	return false;
}

#ifdef Q_OS_LINUX
int functionConversation(int numMsg, const struct pam_message **msg, struct pam_response ** resp, void *appdata_ptr)
{
	Q_UNUSED(appdata_ptr);
	Q_UNUSED(numMsg);
	Q_UNUSED(msg);
	/*for (int i = 0; i < numMsg; i++)
	{
		qDebug("msg: %s", msg[i]->msg);
	}*/
	*resp = reply;

	return PAM_SUCCESS;
}
#endif

bool Session::authenticateSystemUser( QString username, QString password )
{
#ifdef Q_OS_WIN
	//LogonUser(L"LocalServer", L"NT AUTHORITY", NULL, LOGON32_LOGON_SERVICE, LOGON32_PROVIDER_DEFAULT, &hToken)
	return true;
#endif
#ifdef Q_OS_LINUX
	reply = (struct pam_response *)malloc(sizeof(struct pam_response));

	const struct pam_conv local_conversation = { functionConversation, NULL };


	int retval;
	retval = pam_start("check_user", username.toUtf8().data(), &local_conversation, &local_auth_handle);

	if (retval != PAM_SUCCESS)
	{
		return false;
	}

	reply[0].resp = strdup( password.toUtf8().data() );
	reply[0].resp_retcode = 0;
	retval = pam_authenticate(local_auth_handle, 0);
	pam_set_item(local_auth_handle, PAM_RUSER, username.toUtf8().data());

	if (retval != PAM_SUCCESS)
	{
		return false;
	}

	return true;
#endif
}

QString Session::getUserHome( QString username )
{
	QString ret;
#ifdef Q_OS_LINUX
	struct passwd pwd;
	struct passwd *result;
	char *buf;
	size_t bufsize;
	int s;

	bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (bufsize == (size_t)-1)          /* Value was indeterminate */
	{
		bufsize = 16384;        /* Should be more than enough */
	}

	buf = (char*)malloc(bufsize);
	if (buf == NULL)
	{
		qDebug("Error getting memory for buffer");
		return QString::null;
	}

	s = getpwnam_r(username.toUtf8().data(), &pwd, buf, bufsize, &result);
	if (result == NULL)
	{
		if (s == 0)
		{
			qDebug("Not found\n");
		} else {
			qDebug("Error %d on getpwnam_r", s);
		}
	} else {
		ret = QString(result->pw_dir);
	}
	free( buf );
#endif
	return ret;
}

QString Session::createHome( QString homes, QString username )
{
	qDebug("Switching to user %s on %s...", username.toUtf8().data(), homes.toUtf8().data());
	QString homeBase = homes;
	if ( ! homeBase.endsWith(QDir::separator()) )
	{
		homeBase += QDir::separator();
	}
	// Get uid and gid fror username
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	if ( uid == -1 || gid == -1 )
	{
		qDebug("Error obtaining user id/gid");
		return QString::null;
	}

	// Create home dir
	if ( ! QDir(homes).mkdir(username) )
	{
		qDebug("Warning: Cannot create home dir");
	}
#ifdef Q_OS_LINUX
	if ( ! chown( (homeBase + username).toUtf8().data(), uid, gid) != 0 )
	{
		qDebug("Warning: Cannot change home dir owner");
	}
#endif
	return homeBase + username;
}

bool Session::switchUser( QString homeDir, QString username )
{
	// Get uid and gid fror username
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	if ( uid == -1 || gid == -1 )
	{
		qDebug("Error obtaining user id/gid");
		return false;
	}

#ifdef Q_OS_LINUX
	// Create /run/user dir
	if ( ! QDir("/run/user/").mkdir( QString::number(uid) ) )
	{
		qDebug("Warning: Cannot create home dir");
	}
	if ( ! chown( ("/run/user/" + QString::number(uid)).toUtf8().data(), uid, gid) != 0 )
	{
		qDebug("Warning: Cannot change /run/user owner");
	}
	if ( ! chmod( ("/run/user/" + QString::number(uid)).toUtf8().data(), S_IRUSR | S_IWUSR | S_IXUSR) != 0 )
	{
		qDebug("Warning: Cannot change /run/user permissions");
	}
#endif

#ifdef Q_OS_LINUX
	// Get DISPLAY variable
	QByteArray display = qgetenv("DISPLAY");

	// Get number of user groups
	int ngroups = 1;
	gid_t * groups;
	groups = (gid_t*)malloc(ngroups * sizeof (gid_t));
	getgrouplist(username.toUtf8().data(), uid, groups, &ngroups);
	free(groups);

	// Get user group list
	groups = (gid_t*)malloc(ngroups * sizeof (gid_t));
	getgrouplist(username.toUtf8().data(), uid, groups, &ngroups);

	// Set principal group
	if ( setgid(gid) != 0 )
	{
		qDebug("Error: cannot set group id to %d", gid);
	}

	// Set group list
	setgroups(ngroups, groups);
	free(groups);

	// Set user id
	if ( setuid(uid) != 0 )
	{
		qDebug("Error: cannot set user id to %d", uid);
		return false;
	}

	// Open ConsoleKit session and create XDG_SESSION_COOKIE variable
	QByteArray ssid;
	QDBusInterface dbusManager( "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Manager", "org.freedesktop.ConsoleKit.Manager", QDBusConnection::systemBus() );
	if ( dbusManager.isValid() )
	{
		QDBusMessage msg = dbusManager.call( "OpenSession" );
		if ( ! msg.errorName().isEmpty() )
		{
			qDebug("Cannot open session. Error: %s", msg.errorName().toUtf8().data());
		} else {
			QList<QVariant> args = msg.arguments();
			if ( args.count() >= 1 )
			{
				qputenv("XDG_SESSION_COOKIE", args[0].toByteArray() );
				msg = dbusManager.call( "GetSessionForCookie" );
				if ( ! msg.errorName().isEmpty() )
				{
					qDebug("Cannot get session for cookie %s. Error: %s", args[0].toByteArray().data(), msg.errorName().toUtf8().data());
				} else {
					QList<QVariant> sargs = msg.arguments();
					if ( sargs.count() >= 1 )
					{
						ssid = sargs[0].toByteArray();
					} else {
						qDebug("No response from system bus to get session for cookie.");
					}
				}
			} else {
				qDebug("No response from system bus to open session.");
			}
		}
	} else {
		qDebug("Cannot connect to system bus. Cannot open session.");
	}

	// Activate ConsoleKit session
	if ( ! ssid.isEmpty() )
	{
		QDBusInterface dbusSession( "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Seat", "org.freedesktop.ConsoleKit.Seat", QDBusConnection::systemBus() );
		if ( dbusSession.isValid() )
		{
			QDBusMessage msg = dbusSession.call( "ActivateSession", ssid );
			if ( ! msg.errorName().isEmpty() )
			{
				qDebug("Cannot activate session. Error: %s", msg.errorName().toUtf8().data());
			}
		} else {
			qDebug("Cannot connect to system bus. Cannot activate session.");
		}
	}

	// Set environment variables
	QString shell;
	if ( QFile::exists("/bin/bash") )
	{
		shell = "/bin/bash";
	} else {
		shell = "/bin/sh";
	}

	QStringList paths;
	paths += "/usr/local/bin";
	paths += "/usr/bin";
	paths += "/bin";

	qputenv("SHELL", shell.toUtf8());
	qputenv("PATH", paths.join(":").toUtf8());
	qputenv("USER", username.toUtf8());
	qputenv("UID", QString::number( uid ).toUtf8());
	qputenv("HOME", homeDir.toUtf8());
	qputenv("XDG_RUNTIME_DIR", ("/run/user/" + QString::number( uid )).toUtf8());
	qputenv("DISPLAY", display);
	setCurrentDir( homeDir.toUtf8().data());

	QProcess * dbusLaunch = new QProcess();
	dbusLaunch->start("dbus-launch");
	if ( dbusLaunch->waitForStarted() && dbusLaunch->waitForFinished() )
	{
		while ( dbusLaunch->canReadLine() )
		{
			QString line = QString( dbusLaunch->readLine() );
			QString variable = line.section("=", 0, 0);
			QString content = line.right( line.size() - variable.size() - 1);
			qputenv( variable.toUtf8().data(), content.toUtf8() );
		}
	}

	Settings::setUserName( username );
	Settings::setHomePath( homeDir );

	qDebug("Finished user switching.");
	return true;
#endif
	return false;
}

void Session::createDesktopLinks(QString homeDir, QString username, bool setPermissions)
{
	int uid, gid;
	if ( setPermissions )
	{
		getUserAndGroupIds(username, &uid, &gid);
	}

	QStringList folders;
	folders << "Downloads" << "Documents" << "Desktop";
	QDir home( homeDir );
	for ( int i = 0; i < folders.count(); i++ )
	{
		if ( ! QFile::exists(homeDir + QDir::separator() + folders[i]) )
		{
#ifdef Q_OS_LINUX
			bool ret = home.mkdir( folders[i] );
			if ( ret && setPermissions )
			{
				chown( QString(homeDir + QDir::separator() + folders[i]).toUtf8().data(), uid, gid );
			}
#endif
			if ( folders[i] == "Desktop" )
			{
				QStringList links;
				links << "Downloads" << "Documents";
				for ( int i = 0; i < links.count(); i++ )
				{
#ifdef Q_OS_LINUX
					bool ret = QFile::link( homeDir + QDir::separator() + links[i], homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]);

					if ( ret && setPermissions )
					{
						chown( QString(homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]).toUtf8().data(), uid, gid );
					}
#endif
				}
			}
		}
	}
}

bool Session::openSession()
{
#ifdef Q_OS_LINUX
	int retval = pam_open_session(local_auth_handle, PAM_SILENT);

	return (retval == PAM_SUCCESS);
#endif
	return true;
}

bool Session::closeSession()
{
#ifdef Q_OS_LINUX
	int retval = pam_close_session(local_auth_handle, PAM_SILENT);
	retval = pam_end(local_auth_handle, retval);

	return (retval == PAM_SUCCESS);
#endif
	return true;
}

bool Session::connectToServer()
{
	if ( LWClient::isConnected() )
	{
		return true;
	}

	if ( LWClient::connect() )
	{
		return true;
	} else {
		return false;
	}
}

QStringList Session::launch(QString username, QString password, bool localSession, QWidget * parent, Manager *&manager, InitProgress * initProgress)
{
	QString homeDir = QString::null;
	QStringList error;

	if ( ! localSession && Settings::getValue( "Server/LibreworkServerPort" ).toInt() > 0 ) // Auth against Librework server, for PXE
	{
		if ( connectToServer() )
		{
			if ( LWClient::authenticate( username, password ) )
			{
#ifdef Q_OS_LINUX
				bool ok;
				int uid = LWClient::sendMessage( "get_user_uid" ).toInt( &ok );
				if ( ok )
				{
					if ( ! Session::createSystemUser( uid, QStandardPaths::writableLocation( QStandardPaths::TempLocation ), username ) )
					{
						// Cannot create system user
						error << tr("Cannot create user") << tr("Cannot enable user with id %1 on this system").arg(uid);
						return error;
					}

					homeDir = Session::createHome( QStandardPaths::writableLocation( QStandardPaths::TempLocation ), username );
					if ( homeDir.isNull() )
					{
						error << tr("Cannot create home directory")<< tr("Cannot create home dir for user with id %1 on this system").arg(uid);
						return error;
					}
					mountSmbShares( username, password );
				} else {
					// Cannot get system user uid
					error << tr("Cannot identify user") << tr("Cannot identify user %1 on this system").arg(username);
					return error;
				}
#endif
			} else {
				// Bad authentication
				error << tr("Bad authentication") << tr("Either the user name or the password you provided are invalid");
				return error;
			}
		} else {
			// Cannot connect to Librework Server
			error << tr("Bad connection") << tr("Cannot connect to Librework Server");
			return error;
		}
	} else { // No Librework Server, must authenticate against PAM
		if ( ! Session::authenticateSystemUser(username, password ) )
		{
			 // Bad authentication
			error << tr("Bad authentication") << tr("Either the user name or the password you provided are invalid");
			return error;
		} else {
			homeDir = Session::getUserHome(username);
			if ( homeDir.isNull() )
			{
				// Cannot get user home dir
				error << tr("Cannot identify user home") << tr("Cannot identify home of user %1 on this system").arg(username);
				return error;
			}
			bool ret = Session::openSession();
			if ( ret )
			{
				qDebug("Session succesfully opened");
			} else {
				qDebug("Cannot open session");
			}
		}
	}

	if ( homeDir.isNull() )
	{
		// Cannot get user home dir
		error << tr("Cannot identify user home") << tr("Cannot identify home of user %1 on this system").arg(username);
		return error;
	}

	QString oldLastLogin = Settings::getValue("LastLogin");
	Settings::setValue(CONFIG_FILE, "LastLogin", username);

	if ( Session::switchUser( homeDir, username ) )
	{
		Settings::setUserValue("LastLogin", username);


		QString lang = Settings::getUserValue("Language", "").toString();
		if ( ! lang.isEmpty() )
		{
			QTranslator translator;
			if (translator.load("librework-desktop_" + lang))
			{
				qApp->installTranslator(&translator);
			}
		}
		//manager = new Manager( initProgress, username, password, localSession, parent);
		//manager->init();
	} else {
		Settings::setValue(CONFIG_FILE, "LastLogin", oldLastLogin);
		error << tr("Cannot start user session") << tr("Cannot start a session for user %1").arg(username);
		return error;
	}

	return error; // No error
}

QString Session::credentialsFile(QString username, QString password)
{
	QDir tmp( QStandardPaths::writableLocation( QStandardPaths::TempLocation ) );
	tmp.mkdir( ".librework" );
	QString fileName = QStandardPaths::writableLocation( QStandardPaths::TempLocation )
				+ QString(QDir::separator())
				+ ".librework"
				+ QDir::separator()
				+ ".credentials";
	QFile file(fileName);
	file.setPermissions(QFileDevice::ReadOwner | QFileDevice::WriteOwner);
	file.open(QFile::WriteOnly);
	file.write(QString("username=" + username + "\n").toUtf8());
	file.write(QString("password=" + password + "\n").toUtf8());
	file.close();

	return fileName;
}

void Session::mountSmbShares(QString username, QString password)
{
	smbUmountAll();

	QString credentials = credentialsFile( username, password );

	// Mount home dir at .homepath
	QString smbHomePath;
	QString serverAddress = LWClient::sendMessage("get_profiles_server");
	if ( ! serverAddress.isEmpty() )
	{
		smbHomePath = "//" + serverAddress.section(":", 0, 0) + "/" + username;
	} else {
		smbHomePath = "//" + Settings::getValue("ProfilesLocation") + "/" + username;
	}

	QString baseClientPath = QStandardPaths::writableLocation( QStandardPaths::TempLocation ) + QDir::separator();
	QString targetHomePath = baseClientPath + username;
	QDir("/").mkpath( targetHomePath );
	if (QFile::exists( targetHomePath ) && ! smbHomePath.isEmpty())
	{
		if ( smbMount( username, credentials, smbHomePath, targetHomePath ) )
		{
			qputenv("HOME", targetHomePath.toUtf8());
			qputenv("USER", username.toUtf8());
		}
	}

	// Mount shares
	qDebug("Mounting shares...");
	QStringList groups = LWClient::sendMessage("get_redirect_drives").split(",");
	for ( int i = 0; i < groups.count(); i++ )
	{
		if ( ! groups[i].isEmpty() )
		{
			QString clientPath = groups[i].trimmed().section('=', 0, 0).remove("\\").toUpper();
			QString serverPath = groups[i].trimmed().section('=', 1, 1);
			if ( serverPath.startsWith("\\\\") ) // Shared folder
			{
				serverPath.replace("\\", "/");
				QDir("/").mkpath( baseClientPath + clientPath );
				if ( QFile::exists(baseClientPath + clientPath) )
				{
					smbMount( username, credentials, serverPath, baseClientPath + clientPath);
				}
			} else { // Client drive redirect
				// Do nothing
			}
		}
	}

	// Remove credentials file
	QFile credsFile(credentials);
	credsFile.remove();
	credsFile.close();
}

bool Session::smbMount(QString username, QString credentialsFile, QString smbPath, QString targetPath)
{
#ifdef Q_OS_LINUX
	int uid, gid;
	getUserAndGroupIds(username, &uid, &gid);

	QProcess smbMountProcess;
	QString command = "mount.cifs";
	QStringList parameters;
	parameters << "-o";
	parameters << "cache=loose,fsc,actimeo=60,uid=" + QString::number(uid) + "," +
				  "credentials=" + credentialsFile;
	parameters << smbPath;
	parameters << targetPath;

	qDebug("Mounting samba dir: %s %s", command.toUtf8().data(), parameters.join(" ").toUtf8().data());

	smbMountProcess.start( command, parameters );
	if ( smbMountProcess.waitForStarted() )
	{
		smbMountProcess.waitForFinished();
		return ( smbMountProcess.exitCode() == 0 );
	}
#endif
	return false;
}

void Session::smbUmountAll()
{
	QProcess::startDetached("umount -t cifs -a");
}
