#include "tabbar.h"
#include "manager.h"

#include <QDir>

TabBar::TabBar (InitProgress *initProgress, QWidget * parent ) : QWidget( parent,  Qt::FramelessWindowHint )
{
	this->initProgress = initProgress;

#ifdef Q_OS_WIN
	setWindowFlags( this->windowFlags() | Qt::BypassWindowManagerHint);
#endif
	setWindowFlags( this->windowFlags()	| Qt::WindowStaysOnTopHint
						| Qt::CustomizeWindowHint
						| Qt::WindowDoesNotAcceptFocus );
	setAttribute(Qt::WA_AlwaysShowToolTips);
	setAttribute(Qt::WA_X11NetWmWindowTypeDock);
	setFocusPolicy(Qt::NoFocus);

	setWindowTitle(tr("Librework Desktop Bar"));

	QPalette pal(palette());
	pal.setColor(QPalette::Background, QColor( TABCOLOR ));
	setAutoFillBackground(true);
	setPalette(pal);

	layout = new QHBoxLayout;
	layout->setSizeConstraint( QLayout::SetMinimumSize );
	layout->setSpacing(0);
	layout->setContentsMargins(0, 0, 0, 0);

	menusLayout = new QHBoxLayout;
	menusLayout->setSpacing(0);
	menusLayout->setContentsMargins(0, 0, 0, 0);
	menusLayout->setAlignment( Qt::AlignLeft );
	menusLayout->setSizeConstraint( QLayout::SetMinimumSize );

	if ( Settings::getWindowManagerBehaviour() == FULLSCREEN )
	{
		currentThread = new ThreadWidget( nullptr, false );
		currentThread->setFixedHeight( TAB_BUTTON_HEIGHT );
	}

	threadsBar = new QWidget( nullptr );

	threadsLayout = new QHBoxLayout;
	threadsLayout->setSpacing(0);
	threadsLayout->setContentsMargins(0, 0, 0, 0);
	threadsLayout->setAlignment( Qt::AlignLeft );

	threadsBar->setLayout( threadsLayout );

	leftButton = new QPushButton( QIcon( ICON_DIR_64 + QString(QDir::separator()) + "back.png" ), QString(), this );
	leftButton->setFixedSize( QSize(TAB_BUTTON_WIDTH / 2, TAB_BUTTON_HEIGHT) );
	leftButton->hide();
	leftButton->setAccessibleName( "leftButton" );
	connect( leftButton, SIGNAL(clicked()), this, SLOT(leftButtonClicked()));
	rightButton = new QPushButton( QIcon( ICON_DIR_64 + QString(QDir::separator()) + "forward.png" ), QString(), this );
	rightButton->setFixedSize( QSize(TAB_BUTTON_WIDTH / 2, TAB_BUTTON_HEIGHT) );
	rightButton->hide();
	rightButton->setAccessibleName( "rightButton" );
	connect( rightButton, SIGNAL(clicked()), this, SLOT(rightButtonClicked()));
	threadsLayout->addWidget(leftButton);
	threadsLayout->addWidget(rightButton);

	exitLayout = new QHBoxLayout;
	exitLayout->setContentsMargins(0, 0, 0, 0);
	exitLayout->setSpacing(0);

	layout->addLayout( menusLayout, 0 );
	if ( Settings::getWindowManagerBehaviour() == FULLSCREEN )
	{
		layout->addWidget( currentThread, 0 );
	}
	layout->addWidget( threadsBar, 1 );
	layout->addLayout( exitLayout, 0 );

	setLayout( layout );

	if ( Settings::getWindowManagerBehaviour() == FULLSCREEN )
	{
		connect (currentThread, SIGNAL(closeButtonClicked(ThreadButton *)), this, SLOT(closeButtonClicked(ThreadButton *)) );
		connect (currentThread, SIGNAL(killButtonClicked(ThreadButton *)), this, SLOT(killButtonClicked(ThreadButton *)) );
	}

	buttonList = new QList<ThreadButton*>;
	hiddenThreadButtons = 0;
	firstShownThreadButton = 0;

	addMenuButtons();

	QTimer * dateTimer = new QTimer( this );
	connect( dateTimer, SIGNAL(timeout()), this, SLOT(setDate()));
	dateTimer->start(1000);
}

void TabBar::setDate()
{
	QTime time = QTime::currentTime();
	QDate dat = QDate::currentDate();
	date->setText( time.toString("hh:mm") + "\n" + dat.toString("ddd d MMM") );
}

void TabBar::addMenuButtons()
{
	menu = new Menu();
	ThreadButton * menuButton = new ThreadButton(menu, LOGOMENUBUTTON, QPixmap(ICON_DIR_32 + QString(QDir::separator()) + "panel-start.png") , tr("Librework"), tr("Desktop"), (HWND)menu->winId());
	addButton( menuButton );

	desktop = new Desktop( initProgress, nullptr, BACKGROUNDCOLOR );
	ThreadButton * desktopButton = new ThreadButton(desktop, DESKTOPBUTTON, QPixmap(ICON_DIR_32 + QString(QDir::separator()) + "panel-desktop.png") , tr("Librework"), tr("Desktop"), (HWND)desktop->getWindowId());
	addButton( desktopButton );
	connect( desktop, SIGNAL(launchApplication(Application*)), this, SIGNAL(launchApplication(Application*)));
	connect( desktop, SIGNAL(launchApplication()), this, SIGNAL(launchApplication()));
	lastMenuButton = (QWidget*) desktop;
	desktop->initializeDesktopWidgets();
	connect( desktopButton, SIGNAL(clicked()), this, SIGNAL(desktopButtonClicked()));

	configure = new Configure();
	ThreadButton * configureButton = new ThreadButton(configure, MENUBUTTON,
				  QPixmap(ICON_DIR_32 + QString(QDir::separator()) + "panel-config.png"),
				  tr("Librework"), tr("Configuration"), (HWND)configure->winId());
	addButton( configureButton );
	connect( configureButton, SIGNAL(clicked()), this, SLOT(configureButtonClicked()));

	exitButton = new ThreadButton( nullptr, EXITBUTTON,  QPixmap(ICON_DIR_24 + QString(QDir::separator()) + "logoff.png") );
	addButton( exitButton );
	connect(exitButton, SIGNAL(clicked()), this, SLOT(exitAppSlot()));

	desktop->setLogo();

	configure->addApplications();
	configure->hide();

	menu->hide();

	connect( menu, SIGNAL(favoritesChanged()), desktop, SLOT(reloadFavorites()));
}

void TabBar::exitAppSlot()
{
	calendarWidget->hide();
	QThread::msleep( WINMANAGER_TIMER * 2 );
	emit exitApp();
}

void TabBar::configureButtonClicked()
{
	if ( WinInfo::getTopMostWindow() == (HWND)configure->winId() )
	{
		configure->hide();
	} else {
		configure->show();
		WinAction::showWindow( (HWND)configure->winId() );
	}
}

void TabBar::threadButtonClickedSlot(HWND window)
{
	if ( ! WinInfo::isMinimized( window ) && WinManager::getPenultimateTopMostWindow() == window )
	{
		WinAction::minimizeWindow( window );
	} else {
		emit threadButtonClicked( window );
	}
}

void TabBar::updateApps()
{
	Application::deleteApplicationCache();
	Application::updateApplicationCache();
	menu->updateFilters();
	desktop->resize();
}

void TabBar::closeEvent(QCloseEvent * closeEvent)
{
	closeEvent->ignore();
}

TabBar::~TabBar()
{
	desktop->close();
	menu->close();
	configure->close();

	delete desktop;
	delete menu;
	delete configure;
	delete calendarWidget;
	delete date;
}

int TabBar::numThreadButtons()
{
	return threadsLayout->count() - 2;
}

void TabBar::moveEvent(QMoveEvent *event)
{
	QWidget::moveEvent( event );
	//WinAction::locateWindow( (HWND) winId(), 0, Manager::getDesktopSize().height(), Manager::getDesktopSize().width(), TAB_BUTTON_HEIGHT );
}

void TabBar::resizeEvent(QResizeEvent *event)
{
	QWidget::resizeEvent( event );
}

void TabBar::resize()
{
	QRect geometry = QApplication::primaryScreen()->geometry();
	QRect desktopGeometry( geometry.x(), geometry.y(), geometry.width(), geometry.height() - TAB_BUTTON_HEIGHT);
	QRect tabGeometry( geometry.x(), geometry.height() - TAB_BUTTON_HEIGHT, geometry.width(), TAB_BUTTON_HEIGHT);
	desktop->resize( desktopGeometry.size() );
	desktop->setGeometry( desktopGeometry );
	desktop->setFixedSize( desktopGeometry.size() );

	updateContents();

	Settings::logInfo( QString("Setting new geometry - x: %1, y: %2, width: %3, height: %4").arg(0).arg(geometry.height()).arg(geometry.width()).arg(TAB_BUTTON_HEIGHT));
	setGeometry( tabGeometry );
	setFixedSize( tabGeometry.size() );
	setMaximumSize( tabGeometry.size() );

	qApp->processEvents();

	show();

	//WinAction::locateWindow( (HWND)this->winId(), 0, Manager::getDesktopSize().height(), Manager::getDesktopSize().width(), TAB_BUTTON_HEIGHT );
}

void TabBar::updateContents()
{
	// First, calculate currentThread (ThreadWidget) width
	int occupiedSize = menusLayout->count() * TAB_BUTTON_WIDTH /* static menus */
					+ TAB_BUTTON_WIDTH /* exitButton*/
					+ TAB_BUTTON_WIDTH /* left & right */ 
					+ TAB_BUTTON_WIDTH /* at least one thread button */;

	int currentThreadMaximumWidth = QApplication::primaryScreen()->size().width() - occupiedSize;
	if ( currentThreadMaximumWidth > THREAD_WIDGET_WIDTH )
		currentThreadMaximumWidth = THREAD_WIDGET_WIDTH;

	// Hide all thread buttons
	for (int i = 0; i < numThreadButtons(); i++)
		threadsLayout->itemAt( 1 + i )->widget()->hide();

	// Get left space for thread buttons
	int spaceLeft = QApplication::primaryScreen()->size().width()
			- menusLayout->count() * TAB_BUTTON_WIDTH /* static menus */
			- TAB_BUTTON_WIDTH /* exitButton*/
			- currentThreadMaximumWidth /* currentThread width */;

	// Calculate maximum number of buttons on the given space left (c++ integer division always round down)
	int maxButtons = spaceLeft / TAB_BUTTON_WIDTH;

	// Add to currentThread (ThreadWidget) the remaining space
	int remainingSpace = spaceLeft - ( maxButtons * TAB_BUTTON_WIDTH );
	if ( Settings::getWindowManagerBehaviour() == FULLSCREEN )
	{
		currentThread->setFixedSize( currentThreadMaximumWidth + remainingSpace, TAB_BUTTON_HEIGHT );
	}

	// Calculate number of shown and hidden buttons, and the position of the first shown button
	int buttonsShown = 0;
	if ( numThreadButtons() > maxButtons )
	{
		hiddenThreadButtons = numThreadButtons() - maxButtons + 1;
		buttonsShown = maxButtons - 1; // Right and left (scroll) buttons occupy as one button, so substract it
		firstShownThreadButton = hiddenThreadButtons;
	} else {
		hiddenThreadButtons = 0;
		firstShownThreadButton = 0;
		buttonsShown = numThreadButtons();
	}

	// Finally show calculated thread buttons
	if ( hiddenThreadButtons > 0 )
	{
		for (int i = 0; i < numThreadButtons(); i++)
		{
			if ( i >= firstShownThreadButton && i < firstShownThreadButton + buttonsShown )
			{
				qDebug("shown");
				threadsLayout->itemAt( 1 + i )->widget()->show();
			} else {
				qDebug("hidden");
				threadsLayout->itemAt( 1 + i )->widget()->hide();
			}
		}
		showScrollButtons();
	} else { // as ther are no hidden buttons, hide scroll buttons but show everyone else
		hideScrollButtons();
		for (int i = 0; i < numThreadButtons(); i++)
		{
			threadsLayout->itemAt( 1 + i )->widget()->show();
		}
	}
}

void TabBar::removeButton( int index )
{
	ThreadButton * button;

	if (index < buttonList->count())
	{
		button = buttonList->at( index );
	} else {
		return;
	}

	buttonList->removeAt( index );
	threadsLayout->removeWidget( button );
	delete button;

	updateContents();
}

ThreadButton * TabBar::getButton( int index )
{
	if (index >= 0)
		return buttonList->at( index );
	else
		return nullptr;
}

void TabBar::addToThreadsLayout( ThreadButton * button )
{
	threadsLayout->removeWidget( rightButton );
	threadsLayout->addWidget( button );
	threadsLayout->addWidget( rightButton );
}

void TabBar::addButton(ThreadButton * button)
{
	if (button->getType() != EXITBUTTON)
	{
		int index = indexOf( button->getWinId() );
		buttonList->append( button );

		button->setIndex( index );

		connect( button, SIGNAL(threadButtonClicked(HWND)), this, SLOT(threadButtonClickedSlot(HWND)) );
	}

	switch (button->getType())
	{
		case LOGOMENUBUTTON:
			menusLayout->addWidget( button );
			break;
		case MENUBUTTON:
		case DESKTOPBUTTON:
			menusLayout->addWidget( button );
			break;
		case THREADBUTTON:
			addToThreadsLayout( button );
			break;

		case EXITBUTTON:
			SystrayWidget * systrayWidget = new SystrayWidget();
			date = new QPushButton;
			date->setFlat( true );
			date->setStyleSheet("QPushButton { color : white; font-size : 14px; Text-align: center }");
			calendarWidget = new QCalendarWidget;
			calendarWidget->setWindowFlags( Qt::CustomizeWindowHint | Qt::WindowCloseButtonHint );
			calendarWidget->setWindowTitle(tr("Librework Calendar"));
			calendarWidget->setGridVisible( true );
			calendarWidget->setWindowIcon( QIcon(":/icons/64x64/calendar.png") );
			calendarWidget->hide();
			connect( date, SIGNAL(clicked()), this, SLOT(dateClicked()));
			exitLayout->addWidget( systrayWidget );
			exitLayout->addWidget( date );
			exitLayout->addWidget( button );
			break;
	}
}

void TabBar::dateClicked()
{
	if ( calendarWidget->isVisible() && ! calendarWidget->isMinimized() )
	{
		calendarWidget->hide();
	} else {
		QSize calendarSize(420, 270);
		calendarWidget->setFixedSize( calendarSize );
		qApp->processEvents();
		calendarWidget->setGeometry( qApp->primaryScreen()->geometry().width() - calendarSize.width() - 20,
						 qApp->primaryScreen()->geometry().height() -  calendarSize.height() - this->height(),
						 calendarSize.width(),
						 calendarSize.height() );
		calendarWidget->show();
		WinAction::showWindow( (HWND)calendarWidget->winId() );
	}
}

ThreadButton * TabBar::getFirstShownThreadButton()
{
	if (numThreadButtons() > 0)
        	return (ThreadButton*) threadsLayout->itemAt( 1 + firstShownThreadButton )->widget();
	else
		return nullptr;
}

void TabBar::showScrollButtons()
{
	leftButton->show();
	rightButton->show();
	if (firstShownThreadButton == 0)
	{
		leftButton->setEnabled( false );
		rightButton->setEnabled( true );
	} else {
		leftButton->setEnabled( true );
		if (hiddenThreadButtons > firstShownThreadButton)
		{
			rightButton->setEnabled( true );
		} else {
			rightButton->setEnabled( false );
		}
	}
}

void TabBar::hideScrollButtons()
{
	leftButton->hide();
	rightButton->hide();
}

int TabBar::indexOf(HWND window, HWND railWindow)
{
    int index = -1;

    for (int i = 0; i < buttonList->count(); i++)
    {
		if ( railWindow != 0)
		{
			if (buttonList->at(i)->getRailWindow() == (HWND)railWindow)
			{
				index = i;
			}
		} else {
			if (buttonList->at(i)->getWinId() == (HWND)window)
			{
				index = i;
			}
        }
    }
    return index;
}

int TabBar::indexOf(QWidget *widget)
{
    int index = -1;

    for (int i = 0; i < buttonList->count(); i++)
    {
        if (buttonList->at(i)->getWidget() == widget)
        {
            index = i;
        }
    }
    return index;
}

int TabBar::count()
{
	return buttonList->count();
}

int TabBar::threadButtonCount()
{
	return menusLayout->count();
}

int TabBar::menuButtonCount()
{
    int count = 0;
    for (int i = 0; i < buttonList->count(); i++)
    {
		if ( buttonList->at(i)->getType() <= DESKTOPBUTTON )
        {
            count++;
        }
    }
    return count;
}

void TabBar::leftButtonClicked()
{
	if (firstShownThreadButton > 0)
	{
		int visibleButtons = numThreadButtons() - hiddenThreadButtons;
		threadsLayout->itemAt( firstShownThreadButton + visibleButtons)->widget()->hide();
		firstShownThreadButton--;
		getFirstShownThreadButton()->show();
	}

	showScrollButtons();
}

void TabBar::rightButtonClicked()
{
	int visibleButtons = numThreadButtons() - hiddenThreadButtons;
	getFirstShownThreadButton()->hide();
	threadsLayout->itemAt( firstShownThreadButton + visibleButtons + 1)->widget()->show();
	firstShownThreadButton++;
	showScrollButtons();
}

void TabBar::closeButtonClicked(ThreadButton * button)
{
	emit closeButtonClicked( button->getWinId() );
}

void TabBar::killButtonClicked(ThreadButton * button)
{
	emit killButtonClicked( button->getWinId() );
}

QList<HWND> TabBar::getOwnWindows(bool includeOwnWidgets)
{
	QList<HWND> ownWindows;

	ownWindows += (HWND)winId();
	ownWindows += (HWND)menu->winId();
	ownWindows += (HWND)desktop->winId();
	if ( includeOwnWidgets )
	{
		ownWindows += (HWND)configure->winId();
		ownWindows += (HWND)calendarWidget->winId();
	}

	return ownWindows;
}

void TabBar::removeWindow( HWND window )
{
	int ix = indexOf( window );
	if ( ix >= menuButtonCount() )
	{
		removeButton( ix );
	}
}

void TabBar::newWindow( HWND window )
{
	if ( indexOf( window ) == -1 && ! WinManager::getOwnWindows().contains( window ) )
	{
		QString title = WinInfo::getTitle( window );
		QPixmap pixmap = WinInfo::getIcon( window );
		ThreadButton * button = new ThreadButton(nullptr, THREADBUTTON, pixmap, title, title, window);
		addButton( button );
		emit newThread();
	}
}

void TabBar::setSelected( HWND window )
{
	QList<ThreadButton*> * buttonList = getButtonList();
	for ( int i = menuButtonCount(); i < buttonList->count(); i++ )
	{
		bool isSelected = ( (HWND)buttonList->at(i)->getWinId() == window  );
		buttonList->at(i)->setSelected( isSelected );
		buttonList->at(i)->repaint();
	}
}

void TabBar::changeIcon( HWND window, QPixmap pixmap )
{
	int ix = indexOf( window );
	if ( ix != -1 )
	{
		if ( buttonList->at(ix)->getType() == THREADBUTTON )
		{
			buttonList->at(ix)->setPixmap( pixmap );
		}
	}
}
