#include <QApplication>
#include <QFontDatabase>
#include <QFont>
#include <QDir>
#include <QStringList>
#include <QList>
#include <QFile>

#include <iostream>

#include "../common/constants.h"
#include "../common/filesystemwidget.h"
#include "../common/settings.h"
#include "../common/application.h"
#include "../common/power.h"

#ifdef Q_OS_WIN
#include <windows.h>
#include <fcntl.h>
#include <io.h>
#else
#include <termios.h>
#include <unistd.h>
#endif

void setStdinEcho(bool enable = true)
{
#ifdef Q_OS_WIN
	HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE); 
	DWORD mode;
	GetConsoleMode(hStdin, &mode);

	if( ! enable )
	{
		mode &= ~ENABLE_ECHO_INPUT;
	} else {
		mode |= ENABLE_ECHO_INPUT;
	}

	SetConsoleMode(hStdin, mode );
#else
	struct termios tty;
	tcgetattr(STDIN_FILENO, &tty);
	if( !enable )
	{
		tty.c_lflag &= ~ECHO;
	} else {
		tty.c_lflag |= ECHO;
	}
	(void) tcsetattr(STDIN_FILENO, TCSANOW, &tty);
#endif
}

void usage(QString commandName, QString error = QString::null)
{
	if ( ! error.isEmpty() )
	{
		qDebug("Error: %s", error.toUtf8().data());
	}
	qDebug("Usage: %s [ - i ] [ -a authfile ] <URL to open>", commandName.toUtf8().data());
	qDebug("Example: %s -a /tmp/authfile.tmp /home/user", commandName.toUtf8().data());
	exit(1);
}

int main(int argc, char ** argv)
{
	//Power::instance()->suspend();

	QString openLocation;

	QApplication qApplication(argc, argv);

	QDir fontsDir( LIBREWORK_FONT_DIR );
	QStringList fontFile = fontsDir.entryList(QStringList("*.ttf"));
	QList<int> fontId;
	for (int i = 0; i < fontFile.count(); i++)
	{
		fontId.append( QFontDatabase::addApplicationFont( LIBREWORK_FONT_DIR + QString(QDir::separator()) + fontFile.at(i) ) );
	}
	QFontDatabase fontDatabase;
	QString family = QFontDatabase::applicationFontFamilies(fontId.at(0)).at(0);
	QFont font = fontDatabase.font( Settings::getUserValue("Font", LIBREWORK_FONT).toString(), QString::null, Settings::getUserValue("FontSize", QString::number(LIBREWORK_FONT_BASE_SIZE)).toInt());
	qApplication.setFont( font );

	QString password;

	QStringList arguments = QCoreApplication::arguments();

	bool readFromStdin = false;
	int pos = arguments.indexOf("-i", 0); // Read password from stdin option
	if ( pos != -1 )
	{
		arguments.removeAt( pos );
		readFromStdin = true;
	}

	QString authFileString;
	pos = arguments.indexOf("-a", 0); // Read password from file option
	if ( pos != -1 )
	{
		if ( arguments.count() > pos + 1 )
		{
			authFileString = arguments[pos + 1];
			arguments.removeAt( pos + 1 );
			arguments.removeAt( pos );
		} else {
			usage( argv[0] );
		}
	}

	if ( ! authFileString.isEmpty() )
	{
		QFile authFile( authFileString );
		authFile.open( QFile::ReadOnly );
		QByteArray line = authFile.readLine();
		if ( line.size() > 0 )
		{
			Settings::setPassword( line.trimmed() );
		} else {
			authFile.close();
			usage( argv[0], "Empty password on auth file");
		}
		authFile.close();
		authFile.remove();
	}

	if ( readFromStdin )
	{
#ifdef Q_OS_WIN32
		_setmode(_fileno(stdin), _O_BINARY);
#endif
		qDebug("Waiting username,password,serveraddress,serverport from stdin...");
		setStdinEcho( false );

		QStringList stdinData;
		for ( int i = 0; i < 4; i++ )
		{
			bool exitLoop = false;
			while ( ! exitLoop )
			{
				char arr[1];
				std::cin.read(arr,sizeof(arr));
				if ( arr[0] == '\n' )
				{
					exitLoop = true;
				} else {
					stdinData.append( QString(arr[0]) );
				}
			}
		}
		if ( stdinData.count() == 4 )
		{
			Settings::setUserName( stdinData[0] );
			Settings::setPassword( stdinData[1] );
			Settings::setServerAddress( stdinData[2] );
			Settings::setServerPort( stdinData[3].toInt() );
		}

		setStdinEcho( true );
	}

	if ( Settings::getValue( "Server/LibreworkServerPort" ).toInt() > 0 )
	{
		QString token;
		if ( password.isEmpty() )
		{
			qDebug("Looking for token on file: %s", QString(TOKEN_FILE).toUtf8().data());
			if (QFile::exists( TOKEN_FILE ))
			{
				token = Settings::getValue( TOKEN_FILE, "UserToken" );
			}
		} else {
			token = password;
		}

		if ( ! password.isEmpty() )
		{
			if ( LWClient::instance()->connect() )
			{
				if ( LWClient::instance()->authenticate( Settings::getFullUserName(), token ) )
				{
					qDebug("Succesfully authenticated user token on Librework Server");
				} else {
					qDebug("Cannot authenticate on Librework server with user %s and token %s", Settings::getFullUserName().toUtf8().data(), token.toUtf8().data());
				}
			}
		}
	}

	Application::updateApplicationCache();

	Settings::setExtensions();

	if (arguments.count() >= 2)
	{
		openLocation = arguments[1];
	}

	FileSystemWidget * fileSystemWidget = new FileSystemWidget(0, true, openLocation);

	fileSystemWidget->show();

	return qApplication.exec();
}
